
#include "stm32f10x.h"
#include "stdlib.h"
#include "DCF77.h"

// "OPTS" : damit beim Debug-Kompilat erstens "fMHz" reicht und zweitens das Programm noch im Ram (statt Flash) passt
#define OPTS __attribute__((optimize("-Os")))

#define my_assert(Condition, InfoVal)  if(!(Condition) && _State->Error==_ErrOk) _State->Error=InfoVal;


// "Pseudozufallsfolge p(t) mit einem neunstufigen rckgekoppelten Schieberegister"
// - siehe "PTB-Mitteilungen 114 (2004), Heft 4"
static const uint32_t PRNCodes[512/32+1] =
{
  0x6a9c8620, 0x144ecbd8, 0xd24df96a, 0x0332b27e,
  0x7f4b14c6, 0x79a6b8d1, 0xedac177c, 0x0aafad06,
  0x03ba7a94, 0x15d792e7, 0x6de87309, 0x3fe1ee16,
  0x099d1f78, 0x7cf16e52, 0x6c712a36, 0x80446475,
  0x6a9c8620, // PRNCodes[0] wiederholt
};

static TDCF77State*   _State   = 0; // separat zu _Control, damit _Control hier im Source "verborgener" bleibt
static TDCF77Control* _Control = 0;

//################################################### TimedAdc_Init
static int32_t* OPTS TimedAdc_Init(TTimedAdc* s, int32_t ppm)
{
  // - Quarzabweichung ausgleichbar
  // - Grobabgleich per ppm-Parameter in 1 ppm Schritten
  //   (positiver Wert: Quarz schwingt schneller als Nominalfrequenz)
  // - Feinabgleich per TimedAdc_Tweak() in 1/256 ppm Schritten
  // - verursacht ca. +-2 Trgerphasenrauschen bei fCPU=16MHz und fCarrier=77.5kHz (=77.5kHz/16MHz*360)
  // - der DMA darf weiterlaufen
  int32_t fADC = fCarrier*ADCSamplesPerPeriod;
  int n0 = fMHz*(1000*1000+ppm)/fADC; // STM32F030 DIV?
  int a0 = fMHz*(1000*1000+ppm)-fADC*n0;
  my_assert(n0<=254, _ErrTimedAdc);
  int a1 = fADC;
  int w = -a1;
  int ns = 0;
  for(int i=0; i<AdcBufCnt; i++)
  {
    int n = n0;
    w += a0;
    if(w>0)
    {
      n = n0+1;
      w -= a1;
    }
    s->TimArr[i] = n-1;
    ns += n;
  }
  s->n  = n0;
  s->a0 = fMHz*(1000*1000+ppm)*AdcBufCnt-fADC*ns;
  s->a1 = fADC;
  s->w  = -s->a1;
  return &s->d_ppm_L10; // (darf von auen manipuliert werden)
}

//################################################### TimedAdc_Tweak
static void OPTS TimedAdc_Tweak(TTimedAdc* s)
{
  // je TimedAdc.TimArr[]-DMA-Durchlauf _genau_einmal_ aufrufen:
  // Der Feinabgleichsparameter "d_ppm_L10" darf jederzeit von auen
  // gendert werden.
  int n = s->n; // Standard (kein extra CPU-Takt)
  s->w += s->a0;
  if(s->w>0)
  {
    n = s->n+1; // Sonderfall: ein CPU-Takt extra
    s->w -= s->a1;
  }
  s->TimArr[0] = n-1;
  // neuen Korrekturwert anwenden?
  if(s->d_ppm_L10!=0)
  {
    int a0 = s->a0+(s->d_ppm_L10*AdcBufCnt*fMHz)/1024;
    s->d_ppm_L10 = 0;
    int d = 0;
    if(a0<0) d=+1;
    if(a0>=s->a1) d=-1;
    s->n -= d;
    s->a0 = a0+d*s->a1;
  }
}

//################################################### DoDCT
static void OPTS DoDCT(int16_t* Arr, int32_t PeriodCnt, int32_t* ValC, int32_t* ValS)
{
  // berlagerung mit Amplitude von 256
  my_assert(ADCSamplesPerPeriod>=3 && ADCSamplesPerPeriod<=12, _ErrDCT);
  int32_t SumC = *ValC;
  int32_t SumS = *ValS;
  do
  {
    int32_t v;
    // (Code wurde automatisiert erstellt)
    if(ADCSamplesPerPeriod==3)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==4)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 2
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS -= v*256;
    }
    if(ADCSamplesPerPeriod==5)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 2
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 3
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 4
      SumC += v* 79;
      SumS -= v*243;
    }
    if(ADCSamplesPerPeriod==6)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 5
      SumC += v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==7)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*159;
      SumS += v*200;
      v = *Arr++;    // 2
      SumC -= v* 57;
      SumS += v*249;
      v = *Arr++;    // 3
      SumC -= v*230;
      SumS += v*111;
      v = *Arr++;    // 4
      SumC -= v*230;
      SumS -= v*111;
      v = *Arr++;    // 5
      SumC -= v* 57;
      SumS -= v*249;
      v = *Arr++;    // 6
      SumC += v*159;
      SumS -= v*200;
    }
    if(ADCSamplesPerPeriod==8)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*181;
      SumS += v*181;
      v = *Arr++;    // 2
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 3
      SumC -= v*181;
      SumS += v*181;
      v = *Arr++;    // 4
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 5
      SumC -= v*181;
      SumS -= v*181;
      v = *Arr++;    // 6
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 7
      SumC += v*181;
      SumS -= v*181;
    }
    if(ADCSamplesPerPeriod==9)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*196;
      SumS += v*164;
      v = *Arr++;    // 2
      SumC += v* 44;
      SumS += v*252;
      v = *Arr++;    // 3
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 4
      SumC -= v*240;
      SumS += v* 87;
      v = *Arr++;    // 5
      SumC -= v*240;
      SumS -= v* 87;
      v = *Arr++;    // 6
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 7
      SumC += v* 44;
      SumS -= v*252;
      v = *Arr++;    // 8
      SumC += v*196;
      SumS -= v*164;
    }
    if(ADCSamplesPerPeriod==10)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*207;
      SumS += v*150;
      v = *Arr++;    // 2
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 3
      SumC -= v* 79;
      SumS += v*243;
      v = *Arr++;    // 4
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 5
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 6
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 7
      SumC -= v* 79;
      SumS -= v*243;
      v = *Arr++;    // 8
      SumC += v* 79;
      SumS -= v*243;
      v = *Arr++;    // 9
      SumC += v*207;
      SumS -= v*150;
    }
    if(ADCSamplesPerPeriod==11)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*215;
      SumS += v*138;
      v = *Arr++;    // 2
      SumC += v*106;
      SumS += v*232;
      v = *Arr++;    // 3
      SumC -= v* 37;
      SumS += v*253;
      v = *Arr++;    // 4
      SumC -= v*167;
      SumS += v*193;
      v = *Arr++;    // 5
      SumC -= v*245;
      SumS += v* 72;
      v = *Arr++;    // 6
      SumC -= v*245;
      SumS -= v* 72;
      v = *Arr++;    // 7
      SumC -= v*167;
      SumS -= v*193;
      v = *Arr++;    // 8
      SumC -= v* 37;
      SumS -= v*253;
      v = *Arr++;    // 9
      SumC += v*106;
      SumS -= v*232;
      v = *Arr++;    // 10
      SumC += v*215;
      SumS -= v*138;
    }
    if(ADCSamplesPerPeriod==12)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*221;
      SumS += v*127;
      v = *Arr++;    // 2
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 5
      SumC -= v*221;
      SumS += v*127;
      v = *Arr++;    // 6
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 7
      SumC -= v*221;
      SumS -= v*127;
      v = *Arr++;    // 8
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 9
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 10
      SumC += v*128;
      SumS -= v*221;
      v = *Arr++;    // 11
      SumC += v*221;
      SumS -= v*127;
    }
    if(ADCSamplesPerPeriod==8 && 1==2)
    {
      enum
      {
        ScaleSin00 =   0,
        ScaleSin45 = 256,
        ScaleSin90 = 362  // =ScaleSin45/sin(45)
      };
      int v;
      //                    // #   cos   sin
      v = *Arr++;
      SumC += v*ScaleSin90; // 0  +100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC += v*ScaleSin45; // 1   +71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 2     0  +100
      SumS += v*ScaleSin90;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 3   -71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC -= v*ScaleSin90; // 4  -100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 5   -71   -71
      SumS -= v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 6     0  -100
      SumS -= v*ScaleSin90;
      v = *Arr++;
      SumC += v*ScaleSin45; // 7   +71   -71
      SumS -= v*ScaleSin45;
    }
  } while(--PeriodCnt>0);
  *ValC = SumC;
  *ValS = SumS;
}

//############################################################## LogBase2
static inline int32_t __attribute__((always_inline)) LogBase2(uint32_t v)
{
  // "Find the log base 2 of an N-bit integer in O(lg(N)) operations with multiply and lookup"
  // ("https://graphics.stanford.edu/~seander/bithacks.html")
  int r;      // result goes here
  static const int8_t MultiplyDeBruijnBitPosition[32] =
  {
    0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
    8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31
  };
  v |= v >> 1; // first round down to one less than a power of 2
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;
  r = MultiplyDeBruijnBitPosition[(uint32_t)(v * 0x07C4ACDDU) >> 27];
  return r;
}

//################################################### xlog2
static int32_t OPTS xlog2(uint32_t v)
{
  // Nherung von y = ln(v)/ln(2)*65536
  // - Fehler
  //   nTab |  Err
  //   -----+-----
  //     8  |  166
  //    16  |   45
  //    32  |   13
  //    64  |    5
  enum
  {
    nBitTable = 5,
    nTable    = 1<<nBitTable,
    ScaleUp   = 65536>>nBitTable
  };
  int n = LogBase2(v)-16;
  if(n<0) v<<=-n;
  if(n>0) v>>=+n;
  static const uint16_t Tab[nTable] =
  {
    // Tab[i] = int(log(1+1.0*i/nTable)/log(2)*65536);
    0, 2909, 5731, 8472, 11136, 13726, 16248, 18704, 21097, 23432, 25710, 27935,
    30109, 32234, 34312, 36345, 38336, 40285, 42195, 44068, 45904, 47704, 49472,
    51207, 52910, 54584, 56228, 57844, 59433, 60996, 62534, 64047,
  };
  int p = (v>>(16-nBitTable))-nTable;
  int c0 = Tab[p];
  int c1 = p<nTable-1 ? Tab[p+1] : 65536;
  int c = ((v&(ScaleUp-1))*(c1-c0))>>(16-nBitTable);
  return c0+c+(16+n)*65536;
}

//################################################### xexp2
static int32_t OPTS xexp2(int32_t v)
{
  // Nherung von y = exp(v/65536.0*ln(2))
  enum
  {
    nBitTable = 5,
    nTable    = 1<<nBitTable,
  };
  static const uint16_t Tab[nTable] =
  {
    // Tab[i] = int((exp(1.0*(i)/nTable*log(2))-1)*65536);
    0, 1435, 2901, 4400, 5931, 7496, 9095, 10729, 12399, 14106, 15850,
    17632, 19453, 21314, 23216, 25159, 27145, 29175, 31249, 33368, 35534,
    37747, 40009, 42320, 44681, 47095, 49561, 52082, 54657, 57289, 59978,
    62727
  };
  int p = (v&0xffff)>>(16-nBitTable);
  int c0 = Tab[p];
  int c1 = p<nTable-1 ? Tab[p+1] : 65536;
  int e = 65536+c0+(((c1-c0)*(v&((1<<16)/nTable-1)))>>(16-nBitTable));
  int n = (v>>16)-16;
  if(n<0) e>>=-n;
  if(n>0) e<<=+n;
  return e;
}

//############################################################## uint8_atan2
static uint8_t OPTS uint8_atan2(int32_t y, int32_t x)
{
  // Nherung von y = int(Period*atan2(y, x)/2/4/atan(1))&(Period-1)
  // -  +-1 Abweichung bei Period==256 und AtnCnt==32
  enum
  {
    Period = 256,
    AtnShift = 5,
    AtnCnt = 1<<AtnShift
  };
  static const uint8_t AtnTab[AtnCnt] =
  {
    // AtnTab[i] = int(Period*atan(1.0*i/(AtnCnt-1))/2/4/atan(1)+.5);
    0x00, 0x01, 0x03, 0x04, 0x05, 0x07, 0x08, 0x09,
    0x0a, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12,
    0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
    0x1b, 0x1c, 0x1c, 0x1d, 0x1e, 0x1f, 0x1f, 0x20,
  };
  int xAbs = x<0 ? -x : +x;
  int yAbs = y<0 ? -y : +y;
  if(xAbs>0x7fffffff>>AtnShift || xAbs>0x7fffffff>>AtnShift)
  {
    xAbs >>= AtnShift;
    yAbs >>= AtnShift;
  }
  int phi = AtnTab[xAbs<yAbs ? AtnCnt*xAbs/(yAbs+1) : AtnCnt*yAbs/(xAbs+1)]; // STM32F030 DIV?
  if(xAbs<yAbs) phi=Period/4-phi;
  if(x<0) phi=Period/2-phi;
  if(y<0) phi=Period-phi;
  return phi;
}

//############################################################## CalcPhaCoef
static void OPTS CalcPhaCoef(uint8_t* PhaArr,
                             int      PhaCnt,
                             int      Pha16,
                             int      Pha16Step,
                             int*     Coef)
{
  // - Fourier-Koeffizienten bestimmen
  // - nur die Phase wird verarbeitet (Amplitude=1)
  // - PhaArr[]  : Phasenwert von 0 bis 255 (256 entspricht 2*pi)
  // - PhaCnt    : Anzahl der Phasenwerte
  // - Pha16     : um Faktor 65536 vergrerte Startphase der Mischschwingung
  // - Pha16Step : um Faktor 65536 vergrerte Schrittweite der Mischschwingung
  enum
  {
    SinPeriod = 64,   // CosSinTab[]
    Period    = 256   // Phase (0..2*pi) angepasst fr Byte (0..255)
  };
  static const int8_t CosSinTab[64+64/4] =
  {
    // CosSinTab[i] = int(Amplitude*cos(i*2*4*atan(1)/SinPeriod));
     127,  126,  124,  121,  117,  112,  105,   98,
      89,   80,   70,   59,   48,   36,   24,   12,
       0,  -12,  -24,  -36,  -48,  -59,  -70,  -80,
     -89,  -98, -105, -112, -117, -121, -124, -126,
    -127, -126, -124, -121, -117, -112, -105,  -98,
     -89,  -80,  -70,  -59,  -48,  -36,  -24,  -12,
       0,   12,   24,   36,   48,   59,   70,   80,
      89,   98,  105,  112,  117,  121,  124,  126,
     127,  126,  124,  121,  117,  112,  105,   98,
      89,   80,   70,   59,   48,   36,   24,   12,
  };
  int SumC = Coef[0];
  int SumS = Coef[1];
  do
  {
    uint8_t Pha = ((*PhaArr++)+(Pha16>>16))&(Period-1);
    Pha16 += Pha16Step;
    SumC += CosSinTab[(Pha>>2)+0*0x10];
    SumS += CosSinTab[(Pha>>2)+1*0x10];
  } while(--PhaCnt>0);
  Coef[0] = SumC;
  Coef[1] = SumS;
}

//############################################################## Carrier_Init
static void OPTS Carrier_Init(TCarrier* s, int32_t* pd_ppm_L10)
{
  s->pd_ppm_L10 = pd_ppm_L10;
}

//############################################################## CarrierRaw_ProcessSample
static void OPTS Carrier_ProcessSample(TCarrier* s, int32_t ValC, int32_t ValS)
{
  const int RawDownShift = LogBase2((int64_t)SMPPerSecond*Sample_MaxValue/CarrierRawCntPerSecond/0x40000000);
  s->SumC += (ValC-s->PrevC)>>RawDownShift; // mehrfachen berlauf verhindern
  s->SumS += (ValS-s->PrevS)>>RawDownShift;
  s->PrevC = ValC;
  s->PrevS = ValS;
  // Raw
  s->RawSubCur += PeriodsPerSample*CarrierRawCntPerSecond;
  if(s->RawSubCur<fCarrier) return;
  s->RawSubCur -= fCarrier;
  int dC = s->SumC;
  int dS = s->SumS;
  s->SumC = 0;
  s->SumS = 0;
  s->RawPhaArr[s->RawPhaCur] = uint8_atan2(dS, dC);
  s->RawPhaCur++;
  if(s->RawPhaCur>=CarrierRawArrCnt)
  {
    s->RawPhaCur = 0;
  }
  // Fine
  const int FineDownShift = LogBase2((int64_t)SMPPerSecond*Sample_MaxValue/CarrierFineCntPerSecond/0x40000000);
  s->FineSumC += dC>>(FineDownShift-RawDownShift); // mehrfachen berlauf verhindern
  s->FineSumS += dS>>(FineDownShift-RawDownShift);
  s->FineSubCur++;
  if(s->FineSubCur<CarrierRawCntPerSecond/CarrierFineCntPerSecond) return;
  s->FineSubCur = 0;
  s->FinePhaArr[s->FinePhaCur] = uint8_atan2(s->FineSumS, s->FineSumC);
  s->FinePpmArr[s->FinePhaCur] = s->Cur_ppm_L10;
  s->FinePhaCnt++;
  s->FineSumC = 0;
  s->FineSumS = 0;
  s->FinePhaCur++;
  if(s->FinePhaCur>=CarrierFineArrCnt)
  {
    s->FinePhaCur = 0;
  }
  *s->pd_ppm_L10 = s->Next_ppm_L10-s->Cur_ppm_L10; // -> fr TimedAdc_Tweak() anwenden
  s->Cur_ppm_L10 = s->Next_ppm_L10;
}

//############################################################## Carrier_GetResult
static int32_t OPTS Carrier_GetResult(TCarrier* s, bool* InitTimedAdc)
{
  // sollte in der Regel 1x je Sekunde aufgerufen werden
  enum
  {
    Period   = 256,          // Einheit smtlicher Winkelangaben: 2*pi/256
    PhaScale = 65536,

    RawE6Range        = 200, // +-200.0 ppm; Bezug/Mitte: STM32-Quarznominalwert
    RawE6Step         = 4,   // 4.0 ppm Schrittweite
    RawE6FilterConst  = (int64_t)Period*fCarrier*PhaScale/1000000/CarrierRawCntPerSecond,

    FineE7Range       = 100, // +-10.0 ppm; Bezug/Mitte: aktuelle ADC-Frequenz
    FineE7Step        = 1,   // 0.1 ppm Schrittweite
    FineE7FilterConst = (int64_t)Period*fCarrier*PhaScale/1000000/CarrierFineCntPerSecond/10,
    FineE6FilterConst = (int64_t)Period*fCarrier*PhaScale/1000000/CarrierFineCntPerSecond/ 1,

    Next_fADC_Tau_s  = CarrierFineSeconds,

    Sync_Freq_Tau_s  = CarrierFineSeconds,
    Sync_dB_E1_On    = 40,
    Sync_dB_E1_Off   = 20,

  };

  my_assert((CarrierRawCntPerSecond%CarrierFineCntPerSecond)==0, _ErrCarrier_RawCnt);

  // grobe Abweichung bestimmen (ber die vergangene Sekunde)
  if(s->FinePhaCnt>=CarrierFineCntPerSecond)
  {
    int CurHSE_ppm = s->Cur_ppm_L0+s->Cur_ppm_L10/1024;
    int MaxV = 0;
    int MaxP = 0;
    int StateP  = (RawE6Range-CurHSE_ppm)/RawE6Step;
    int StateVS = 0;
    int StateVN = 0;
    int ArrOfs = s->RawPhaCur-CarrierRawCntPerSecond*CarrierRawSeconds;
    if(ArrOfs<0) ArrOfs+=CarrierRawArrCnt;
    int ArrCnt = CarrierRawCntPerSecond*CarrierRawSeconds;
    if(ArrOfs+ArrCnt>CarrierRawArrCnt)
    {
      ArrCnt = CarrierRawArrCnt-ArrOfs;
    }
    int PhaScaledStep = -(RawE6Range-CurHSE_ppm)*RawE6FilterConst;
    for(int P=0; P<2*RawE6Range/RawE6Step+1; P++)
    {
      int Coef[2] = { 0, 0 };
      CalcPhaCoef(s->RawPhaArr+ArrOfs,
                  ArrCnt,
                  0,
                  PhaScaledStep,
                  Coef);
      if(ArrCnt!=CarrierRawCntPerSecond*CarrierRawSeconds)
      {
        CalcPhaCoef(s->RawPhaArr,
                    CarrierRawCntPerSecond*CarrierRawSeconds-ArrCnt,
                    PhaScaledStep*ArrCnt,
                    PhaScaledStep,
                    Coef);
      }
      PhaScaledStep += RawE6Step*RawE6FilterConst;
      int V = Coef[0]*Coef[0]+Coef[1]*Coef[1];
      if(V>MaxV)
      {
        MaxV = V;
        MaxP = P;
      }
      if(P>=StateP-3 && P<=StateP+3)
          {
            if(V>StateVS) StateVS=V;
          }
        else
          {
            if(V>StateVN) StateVN=V;
          }
      #ifdef WIN32
      if(1==2)
      {
        // Debug: Spektrum anzeigen
        int X0 = 0*256+P;
        int Y0 = 70+0*130;
        int y = int(sqrt(V)/CarrierRawCntPerSecond-1);
        if(y>127) y=127;
        static HDC PnDC = GetDC(0);
        BitBlt(PnDC, X0, Y0, 1, 130, NULL, 0, 0, WHITENESS);
        MoveToEx(PnDC, X0, Y0+128, NULL);
        LineTo(PnDC, X0, Y0+128-y);
      }
      #endif
    }
    // Ausreierfilter der mit der Zeit immer strenger wird
    s->RawN++;
    if(MaxP==s->RawP-1 && s->RawN1==0)
    {
      s->RawP--;
      s->RawN1++;
    }
    if(MaxP!=s->RawP)
    {
      s->RawN1++;
      if(MaxP!=s->RawP+1)
      {
        s->RawP = MaxP;
        s->RawN = 1;
        s->RawN1 = 0;
      }
    }
    if(s->RawN==(1<<s->RawNE))
    {
      int RawE6 = (((s->RawP*s->RawN+s->RawN1)*RawE6Step)>>s->RawNE)-RawE6Range;
      if(s->RawNE<5) s->RawNE++;
      int d = CurHSE_ppm+RawE6;
      if(abs(d)>=90*FineE7Range/1000) // auerhalb FineRange?
      {
        s->Cur_ppm_L0 = -RawE6;
        s->RawN = 0;
        s->RawN1 = 0;
        s->FinePhaCnt = 0; // alle Fine-Messwerte verwerfen
        s->Cur_ppm_L10 = 0;
        s->Next_ppm_L10 = 0;
        *InitTimedAdc = true;
      }
    }
    // Signalstrke als "State.Carrier_SNR_Raw_dB" ablegen
    {
      int SNR_xlog2 = (xlog2(StateVS)-xlog2(StateVN))/2;
      _State->Carrier_SNR_Raw_dB = SNR_xlog2*((int)(10*log(2)/log(10)*1))/65536;
    }
  }
  // genaue Abweichung bestimmen (ber 1s..10s)
  // - die jeweilige ADC-Frequenz (FinePpmArr[]) ber die Messzeit wird bercksichtigt
  if(s->FinePhaCnt>=CarrierFineCntPerSecond-1)
  {
    int MaxV = 0;
    int MaxP = 0;
    int MaxCC  = 0;
    int MaxCS  = 0;
    int MaxPha16 = 0;
    int MinV = 0;
    int ArrCnt = s->FinePhaCnt-1; // -1 -> ersten Wert verwerfen
    if(ArrCnt>CarrierFineSeconds*CarrierFineCntPerSecond)
    {
      ArrCnt = CarrierFineSeconds*CarrierFineCntPerSecond;
    }
    int ArrOfs = s->FinePhaCur-ArrCnt;
    if(ArrOfs<0) ArrOfs+=CarrierFineArrCnt;
    my_assert((int64_t)10000*(10*FineE7FilterConst/4)<0x7fffffff, 123);
    int PhaScaledStep = -FineE7Range*FineE7FilterConst
                        -s->Cur_ppm_L10*(FineE6FilterConst/4)/(1024/4);
    for(int P=0; P<2*FineE7Range/FineE7Step+1; P++)
    {
      int Coef[2] = { 0, 0 };
      int Pha16 = 0;
      int Ofs = ArrOfs;
      for(int i=0; i<ArrCnt; )
      {
        int pha = s->FinePpmArr[Ofs];
        int n = Ofs+1;
        for(; n<CarrierFineArrCnt; n++)
        {
          if(s->FinePpmArr[n]!=pha) break;
        }
        n -= Ofs;
        if(i+n>ArrCnt) n=ArrCnt-i;
        i += n;
        int dpha = pha*(FineE6FilterConst/4)/(1024/4);
        CalcPhaCoef(s->FinePhaArr+Ofs,
                    n,
                    Pha16,
                    PhaScaledStep+dpha,
                    Coef);
        Pha16 += (PhaScaledStep+dpha)*n;
        Ofs += n;
        if(Ofs>=CarrierFineArrCnt) Ofs=0;
      }
      PhaScaledStep += FineE7Step*FineE7FilterConst;
      int V = Coef[0]*Coef[0]+Coef[1]*Coef[1];
      if(V>MaxV)
      {
        MaxV = V;
        MaxP = P;
        MaxCC = Coef[0];
        MaxCS = Coef[1];
        MaxPha16 = Pha16;
      }
      if(P>=MaxP+120/CarrierFineSeconds && V>MinV) MinV=V;
      #ifdef WIN32
      if(1==2)
      {
        // Debug: Spektrum anzeigen
        int X0 = 1*256+P;
        int Y0 = 70+0*130;
        int y = int(sqrt(V)/ArrCnt);
        if(y>127) y=127;
        static HDC PnDC = GetDC(0);
        BitBlt(PnDC, X0, Y0, 1, 130, NULL, 0, 0, WHITENESS);
        MoveToEx(PnDC, X0, Y0+128, NULL);
        LineTo(PnDC, X0, Y0+128-y);
      }
      #endif
    }

    // Signalstrke "State.Carrier_SNR_Fine_DB_E1" aktualisieren
    {
      int SNR_xlog2 = (xlog2(MaxV)-xlog2(MinV))/2; // "/2" = "Wurzel aus"
      _State->Carrier_SNR_Fine_dB_E1 = SNR_xlog2*((int)(10*log(2)/log(10)*10))/65536;
    }

    // fADC-Abweichung aus Spektrummaximum berechnen
    int Next_ppm_L10 = s->Cur_ppm_L10 + 1024*(FineE7Range-MaxP*FineE7Step)/10;

    // Anzahl neuer Fine-Werte seit vorherigem Aufruf berechnen
    int32_t nNew = (ArrCnt+ArrOfs)-s->Sync_FinePhaOfs;
    s->Sync_FinePhaOfs += nNew;
    if(nNew<0) nNew+=CarrierFineArrCnt;

    int32_t dphi; // Phasennderung zum vorherigen Aufruf

    // Phasenabweichung fADC zum Trger berechnen
    if(CarrierDoCalc_dCycles!=0 || CarrierDoSynchronize!=0)
    {
      PhaScaledStep -= (2*FineE7Range/FineE7Step+1-MaxP)*FineE7FilterConst;
      int phi0 = (uint8_atan2(MaxCS, MaxCC));
      int dphi0 = (MaxPha16-PhaScaledStep*(ArrCnt/2))/65536; // Intervallende und ohne dCycles_L8 - wird fr Sync gebraucht
      int dphi1 = (MaxPha16/2)/65536; // Intervallmitte (5s alt) (und normal) wird fr "Carrier_dCycles_L8" gebraucht
      dphi = (int8_t)(-(phi0+dphi0)-s->dCycles_L8); // Phasennderung ohne berlufe
      int dE7 = FineE7Range-MaxP*FineE7Step;
      int dphi_Raw = (int32_t)(fCarrier*256*1e-7/CarrierFineCntPerSecond*1024)*dE7*nNew/1024;
      dphi -= ((128+dphi-dphi_Raw)>>8)<<8; // Phasennderung inkl. berlufe
      s->dCycles_L8 += dphi;
      _State->Carrier_dCycles_L8 = s->dCycles_L8+dphi0-dphi1;
    }

    // ADC mit Trgersignal synchronisieren
    if(CarrierDoSynchronize!=0)
    {
      if(_State->Carrier_Synchronized==0)
      {
        if(_State->Carrier_SNR_Fine_dB_E1>=Sync_dB_E1_On &&
           ArrCnt>=CarrierFineSeconds*CarrierFineCntPerSecond)
        {
          s->Sync_ppm_L10_Avr = Next_ppm_L10*1024;
          s->dCycles_L8 += (int8_t)_State->Carrier_dCycles_L8-_State->Carrier_dCycles_L8;
          _State->Carrier_Synchronized = _State->SMPLoop;
        }
      }
      // "den Wert _State->Carrier_dCycles_L8 auf Null regeln"
      if(_State->Carrier_Synchronized!=0)
      {
        // die mittlere dphi-nderung pro Sekunde "Sync_ppm_L10_Avr" ist
        // die Frequenzabweichung
        // (d_ppm_L10 ist Frequenzabweichung seit letzter Messung)
        int d_ppm_L10 = dphi*CarrierFineCntPerSecond*(1024*1000000/fCarrier)/256;
        // die Regelung rausrechnen (-> Bezug auf Cur_ppm_L0)
        for(int i=ArrCnt-nNew; i<ArrCnt; i++)
        {
          int p = ArrOfs+i;
          if(p>=CarrierFineArrCnt) p-=CarrierFineArrCnt;
          d_ppm_L10 += s->FinePpmArr[p];
        }
        // Phasenabweichung zum Nullpunkt (nach und nach) ausgleichen
        {
          const int dftMax = 100; // max. erlaubte Frequenznderung pro Sekunde (HitVal-Qualitt leidet sonst) in 1/1000ppm
          const int cMagic = 17;  // Wert fr schnellstmgliche Regelung ohne berschwinger
          int phi = _State->Carrier_dCycles_L8;
          int v = cMagic*phi/(Next_fADC_Tau_s+Sync_Freq_Tau_s);
          if(v<-dftMax) v=-dftMax;
          if(v>+dftMax) v=+dftMax;
          d_ppm_L10 += v*nNew;
        }
        s->Sync_ppm_L10_Avr += (d_ppm_L10*1024-s->Sync_ppm_L10_Avr*nNew)/(Sync_Freq_Tau_s*CarrierFineCntPerSecond);
        Next_ppm_L10 = s->Sync_ppm_L10_Avr/1024;
        if(_State->Carrier_SNR_Fine_dB_E1<Sync_dB_E1_Off)
        {
          _State->Carrier_Synchronized = 0;
        }
      }
    }

    // neue fADC-Abweichung ber Tiefpass anwenden
    if(ArrCnt<CarrierFineSeconds*CarrierFineCntPerSecond && !s->AlwaysSmooth)
    {
      s->Next_ppm_L10_Avr = Next_ppm_L10*1024;
    }
    s->Next_ppm_L10_Avr += (Next_ppm_L10*1024-s->Next_ppm_L10_Avr)*nNew/(Next_fADC_Tau_s*CarrierFineCntPerSecond);
    s->Next_ppm_L10 = s->Next_ppm_L10_Avr/1024; // "engage" (sichtbar fr IRQ-Routine Carrier_ProcessSample())

  }
  _State->Carrier_dfHSE_E9 = s->Cur_ppm_L0*1000+1000*s->Next_ppm_L10/1024;
  return s->Cur_ppm_L0*1024+s->Next_ppm_L10;
}

//############################################################## DCTAvr512_ProcessSample
static void OPTS DCTAvr512_ProcessSample(TDCTAvr512* s, int32_t ValC, int32_t ValS)
{
  // (Wenn man hier mit 64 Bit-Datentypen rechnen wrde, dann wre alles trivial.)
  enum
  {
    DCFCount = SMPPerSecond/DCTAvr512CntPerSecond,
    MaxCount = 0x7fffffff/Sample_MaxValue,
    SubCount = (DCFCount+MaxCount-1)/MaxCount
  };
  s->Cnt1 += PeriodsPerSample*DCTAvr512CntPerSecond*SubCount;
  if(s->Cnt1<fCarrier) return;
  s->Cnt1 -= fCarrier;
  const int DownShift = LogBase2((int64_t)SMPPerSecond*Sample_MaxValue/0x40000000);
  int dC = (ValC-s->PrevC)>>DownShift; // mehrfachen berlauf verhindern
  int dS = (ValS-s->PrevS)>>DownShift;
  s->PrevC = ValC;
  s->PrevS = ValS;
  s->SumC += dC;
  s->SumS += dS;
  s->Cnt2++;
  if(s->Cnt2<SubCount) return;
  s->Cnt2 = 0;
  if(s->ResetCnt==0)
      {
        s->AvrC = s->SumC-s->ArrC[s->Ofs];
        s->AvrS = s->SumS-s->ArrS[s->Ofs];
      }
    else
      {
        s->ResetCnt--;
        s->AvrC = s->SumC;
        s->AvrS = s->SumS;
      }
  s->ArrC[s->Ofs] = s->SumC;
  s->ArrS[s->Ofs] = s->SumS;
  s->Ofs++;
  if(s->Ofs==DCTAvr512Cnt)
  {
    s->Ofs = 0;
  }
}

//############################################################## DCTAvr512_Reset
static void OPTS DCTAvr512_Reset(TDCTAvr512* s)
{
  s->SumC = 0;
  s->SumS = 0;
  s->ResetCnt = DCTAvr512Cnt;
}

//############################################################## DCTAvr512_GetResult
static void OPTS DCTAvr512_GetResult(TDCTAvr512* s, int32_t* ValC, int32_t* ValS)
{
  *ValC = s->AvrC;
  *ValS = s->AvrS;
}

//############################################################## DCTAvr512_CalcUAdc
static int32_t OPTS DCTAvr512_CalcUAdc(TDCTAvr512* s)
{
  int ValC = s->AvrC;
  int ValS = s->AvrS;
  int Max = abs(ValC)>abs(ValS) ? abs(ValC) : abs(ValS);
  int nBits = LogBase2(Max)+1;
  int nShift = nBits<15 ? 0 : nBits-15;
  ValC >>= nShift;
  ValS >>= nShift;
  int Amp2 = ValC*ValC+ValS*ValS;
  int Amp_xlog2 = xlog2(Amp2)/2+nShift*65536; // "/2" = "Wurzel"
  // DownShift in DCTAvr512_ProcessSample() bercksichtigen
  const int DownShift = LogBase2((int64_t)2*SMPPerSecond*Sample_MaxValue/0x7fffffff-1);
  Amp_xlog2 += DownShift*65536;
  // Bezug auf Eingangspannung mit Sinus Ueff=1V umrechnen
  // (Sample_MaxValue bezieht sich auf Rechteck mit Uss=VCC)
  Amp_xlog2 += (int)(log(1.0*(4/M_PI)*(3.3/2/sqrt(2))/SMPPerSecond/Sample_MaxValue)/log(2)*65536);
  return xexp2(Amp_xlog2+(int)(log(1e6)/log(2)*65536));
}

//############################################################## DCTAvr001_ProcessSample
static void OPTS DCTAvr001_ProcessSample(TDCTAvr001* s, int32_t ValC, int32_t ValS)
{
  const int nCoeff = sizeof(DCTAvr001_Coeff)/sizeof(*DCTAvr001_Coeff);
  s->ArrC[s->Ofs] = (ValC-s->PrevC)>>8;
  s->ArrS[s->Ofs] = (ValS-s->PrevS)>>8;
  s->PrevC = ValC;
  s->PrevS = ValS;
  s->Ofs++;
  if(s->Ofs==nCoeff) s->Ofs=0;
}

//############################################################## DCTAvr001_GetResult
static void OPTS DCTAvr001_GetResult(TDCTAvr001* s, int32_t* ValC, int32_t* ValS)
{
  const int nCoeff = sizeof(DCTAvr001_Coeff)/sizeof(*DCTAvr001_Coeff);
  my_assert(Avr001_SamplesPerPRNBit==SamplesPerPRNBit, _ErrAvr001Coeff);
  int SumC = 0;
  int SumS = 0;
  my_assert(nCoeff<0x7fffffff/Sample_MaxValue, 124); // SumC- bzw. SumS-berlufe
  int p = s->Ofs;
  for(int i=0; i<nCoeff; i++)
  {
    p--;
    if(p<0) p=nCoeff-1;
    SumC += s->ArrC[i]*DCTAvr001_Coeff[p];
    SumS += s->ArrS[i]*DCTAvr001_Coeff[p];
  }
  *ValC = SumC;
  *ValS = SumS;
}

//################################################### PRNBitTst_ProcessBit
static int OPTS PRNBitTst_ProcessBit(TPRNBitTst* s, int Bit)
{
  // das ermittelte PRNBit im Ringpuffer speichern
  {
    if(Bit!=0)
        s->Arr[s->P][(s->I)/32] |= 1<<(s->I&0x1f);
      else
        s->Arr[s->P][(s->I)/32] &= ~(1<<(s->I&0x1f));
  }
  // "PRNCodes[]" berlagern und Trefferzahl bestimmen
  // -256 : DCFBit=0
  //  ... : DCFBit mehr oder weniger 1 oder 0
  // +255 : DCFBit=1
  int HitVal = -256;
  int d = (-s->I-1)&0x1f;
  int p = (-s->I-1)>>5;
  uint32_t* Bits = s->Arr[s->P];
  for(int o=0; o<512/32; o++)
  {
    uint32_t v = Bits[o]; // gemessene Werte
    if(d==0)
        {
          v ^= PRNCodes[(o+p+0)&0x0f];
        }
      else
        {
          v ^= PRNCodes[(o+p+0)&0x0f]>>d;
          v ^= PRNCodes[(o+p+1)&0x0f]<<(32-d);
        }
    // Treffer (=Einser-Bits) in v berechnen
    // ("https://graphics.stanford.edu/~seander/bithacks.html")
    v = v - ((v >> 1) & 0x55555555);                    // reuse input as temporary
    v = (v & 0x33333333) + ((v >> 2) & 0x33333333);     // temp
    HitVal += (((v + (v >> 4)) & 0xF0F0F0F) * 0x1010101) >> 24;
  }
  s->P++;
  if(s->P==SamplesPerPRNBit)
  {
    s->P = 0;
    s->I++;
    if(s->I==512)
    {
      s->I = 0;
    }
  }
  return HitVal;
}

//################################################### PRNBitOfs_ProcessBit
static void OPTS PRNBitOfs_ProcessBit(TPRNBitOfs* s, int32_t Pos, int32_t Val)
{
  if(Val<-0x7f) Val=-0x7f;
  if(Val>+0x7f) Val=+0x7f;
  s->HitAvr -= s->HitArr[s->HitOfs];
  s->HitArr[s->HitOfs] = Val;
  s->HitAvr += s->HitArr[s->HitOfs];
  s->HitOfs++;
  if(s->HitOfs==PRNBitOfs_HitArrCnt) s->HitOfs=0;
  int AvrVal = abs(s->HitAvr);
  // Probe
  for(int i=0; i<PRNBitOfs_ProbeCnt; i++)
  {
    if(Pos==s->ProbePos[i]) s->ProbeVal[i]+=AvrVal*256;
  }
  // Max
  if(AvrVal>s->MaxVal)
  {
    s->MaxVal = AvrVal;
    s->MaxPos = Pos;
  }
  // Focus
  Val = abs(Val);
  int dp = 256*SMPPerSecond+256*Pos-s->FocusPos/256;
  while(dp>256*SMPPerSecond/2) dp-=256*SMPPerSecond;
  if(dp>=-256*PRNBitOfs_dFocusPos && dp<=+256*PRNBitOfs_dFocusPos)
  {
    s->FocusPos += PRNBitOfs_FocusTrack*Val*dp/65536;
  }
}

//############################################################## PRNBitOfs_Reset
static void OPTS PRNBitOfs_Reset(TPRNBitOfs* s)
{
  s->FocusPos = -SMPPerSecond*65536; // -> Reset erzwingen
  for(int i=0; i<PRNBitOfs_ProbeCnt; i++)
  {
    s->ProbeVal[i] = 0;
    s->ProbePos[i] = -1;
  }
}

//############################################################## PRNBitOfs_GetResult
static int OPTS PRNBitOfs_GetResult(TPRNBitOfs* s)
{
  // - zurckgegeben wird der SMPOfs mit maximalem Trefferwert
  //   Ein Alternativalgorithmus per Histogramm wre besser (einfach,
  //   schnell, sicher), aber ein STM32F030 (4KB RAM) hat zu wenig Platz.
  // - aus SMPOfs kann der DCF77-Sekundenbeginn bestimmt werden:
  //   Der maximale Trefferwert liegt beim letzten PRN-Sample des
  //   letzten PRNBits (#511). Das erste PRNBit (#0) beginnt 0.2s nach
  //   DCF77-Sekundenbeginn. Die gesamte Verschiebung der Trefferzahl
  //   hinter Sekundenbeginn betrgt:
  //     dSMPOfs = 2*SMPPerSecond/10 + 512*SamplesPerPRNBit - 1
  //   (Zuzglich der Verzgerung fr eine TDCTAvr001-Fensterfunktion mit
  //   mehr Koeffizienten als SamplesPerPRNBit.)
  //   Beispiel:
  //   Wenn PRNBitOfs_GetResult() 3200 liefert, dann begann die
  //   DCF-Sekunde als "DCF77Control.DCF_SMPOfs" den Wert
  //   3200-(2*SMPPerSecond/10+512*SamplesPerPRNBit-1) hatte.

  // Probe
  int iMin = 0;
  int iMax = 0;
  for(int i=0; i<PRNBitOfs_ProbeCnt; i++)
  {
    if(s->ProbeVal[i]<s->ProbeVal[iMin]) iMin=i;
    if(s->ProbeVal[i]>s->ProbeVal[iMax]) iMax=i;
  }
  int vSub = s->ProbeVal[iMin];
  if(s->ProbeVal[iMax]>PRNBitOfs_ProbeLimit)
  {
    // "Limit" nur, um lange Wartezeiten im Fehlerfall (z.B. wenn einige
    // IRQs verloren gegangen sind) oder Extremfllen (wenn Empfang
    // sehr lange nicht mglich war -> FocusPos durch Drift verloren)
    // zu verhindern.
    vSub = s->ProbeVal[iMax]-PRNBitOfs_ProbeLimit;
  }
  for(int t=0; t<PRNBitOfs_ProbeCnt; t++)
  {
    s->ProbeVal[t] -= vSub;
    if(s->ProbeVal[t]<0) s->ProbeVal[t]=0;
    if(s->ProbePos[t]==s->MaxPos) s->MaxPos=-1;
  }
  s->ProbePos[iMin] = s->MaxPos;
  s->ProbeVal[iMin] = 256*s->MaxVal/PRNBitOfs_ProbeInit;
  // Focus
  int pMax = s->ProbePos[iMax]-PRNBitOfs_HitArrCnt/2;
  int dp = SMPPerSecond+s->FocusPos/65536-pMax;
  while(dp>SMPPerSecond/2) dp-=SMPPerSecond;
  if(abs(dp)>=SamplesPerPRNBit/2)
  {
    s->FocusPos = pMax*65536;
  }
  s->FocusPos += 65536*SMPPerSecond;
  while(s->FocusPos>=65536*SMPPerSecond) s->FocusPos-=65536*SMPPerSecond;
  // Max
  s->MaxPos = 0;
  s->MaxVal = 0;

  int p = (s->FocusPos+32768)/65536;
  if(p>=SMPPerSecond) p-=SMPPerSecond;
  return p;
}

//############################################################## IRQ_ProcessSample
static void OPTS IRQ_ProcessSample(TDCF77Control* s, int32_t CurC, int32_t CurS)
{
  // HSE-Frequenz bestimmen
  Carrier_ProcessSample(&s->Carrier, CurC, CurS);

  // groe fADC-nderung?
  if(s->OscHop)
  {
    s->OscHop = false;
    DCTAvr512_Reset(&s->DCTAvr512);
    PRNBitOfs_Reset(&s->PRNBitOfs);
  }

  // DCT Koeffizient aller PRNBits bestimmen
  // (Eigentlich ber eine Sekunde statt ber die 512Bits (ca. 793 ms).)
  DCTAvr512_ProcessSample(&s->DCTAvr512, CurC, CurS);
  int32_t Avr512C = 0;
  int32_t Avr512S = 0;
  DCTAvr512_GetResult(&s->DCTAvr512, &Avr512C, &Avr512S);

  // DCT Koeffizient vom aktuellen PRNBit bestimmen
  DCTAvr001_ProcessSample(&s->DCTAvr001, CurC, CurS);
  int32_t Avr001C = 0;
  int32_t Avr001S = 0;
  DCTAvr001_GetResult(&s->DCTAvr001, &Avr001C, &Avr001S);

  // Wert vom PRNBit bestimmen
  // (per Auswertung der Phase beim Bit bezglich der mittleren Phase)
  int64_t PhaseAmp = (int64_t)Avr001C*-Avr512S+(int64_t)Avr001S*+Avr512C; // (Kreuzprodukt)
  int32_t PRNBit = PhaseAmp<0 ? 1 : 0;

  // Trefferwert ermitteln (512 PRNBits mit PRNCodes korrelieren)
  int HitVal = PRNBitTst_ProcessBit(&s->PRNBitTst, PRNBit);

  // Startbit suchen
  PRNBitOfs_ProcessBit(&s->PRNBitOfs, s->SMPOfs,  HitVal);

  // DCF-Bit bestimmen und ablegen
  {
    int dp = SMPPerSecond+s->SMPOfs-s->DCF_SMPOfs;
    while(dp>+SMPPerSecond/2) dp-=SMPPerSecond;
    if(dp>=-SamplesPerPRNBit && dp<=+SamplesPerPRNBit-1)
    {
      if(dp==0)
      {
        s->DCF_HitVal = HitVal;
      }
      if(dp>=-DCF77Control_dpAvr && dp<=+DCF77Control_dpAvr)
      {
        s->DCF_HitSum += HitVal;
      }
      if(dp==SamplesPerPRNBit-1)
      {
        // aus SMPOfs vom letzten PRNCodes-Bit (="s->DCF_SMPOfs")
        // den SMPOfs fr Sekundenbeginn berechnen
        const int dSMP = +((int32_t)(sizeof(DCTAvr001_Coeff)/sizeof(*DCTAvr001_Coeff))-SamplesPerPRNBit)/2;
        int SMPOfs = s->DCF_SMPOfs-(2*SMPPerSecond/10+512*SamplesPerPRNBit-1+dSMP);
        if(SMPOfs<0) SMPOfs+=SMPPerSecond;
        _State->DCF77_SMPOfs = SMPOfs;
        _State->DCF77_BitCnt++;
        _State->DCF77_HitVal = s->DCF_HitVal;
        _State->DCF77_BitVal = s->DCF_HitSum>0 ? 1 : 0;
        _State->DCF77_TotalBitCnt++;
        s->DCF_HitSum = 0;
        s->DCF_AvrVal += (256/DCF77Control_sAvrTau)*(256*abs(s->DCF_HitVal)-s->DCF_AvrVal)/256;
        _State->DCF77_AvrAbsHitVal = s->DCF_AvrVal/256;
        // Signalstrke "DCF77State.DCTAvr512_Ueff_uV" aktualisieren
        _State->Signal_Ueff_uV = DCTAvr512_CalcUAdc(&s->DCTAvr512);
      }
    }
    if(s->DCF_SMPOfs!=s->DCF_SMPOfs_Next && abs(dp)>=1+SamplesPerPRNBit)
    {
      // DCF_SMPOfs_Next als nchsten DCF_SMPOfs bernehmen
      int dp = SMPPerSecond+s->DCF_SMPOfs-s->DCF_SMPOfs_Next;
      while(dp>SMPPerSecond/2) dp-=SMPPerSecond;
      // Reset?
      if(abs(dp)>SamplesPerPRNBit)
      {
        // Eigentlich drfte die Schwelle auch viel hher sein. Wichtig ist,
        // dass nie eine Sekunde eingefgt oder verloren geht.
        _State->DCF77_BitCnt = 0;
        _State->DCF77_RstCnt++;
      }
      s->DCF_SMPOfs = s->DCF_SMPOfs_Next;
    }
  }

  // nchster Durchgang?
  // - Bei SamplesPerPRNBit!=6 entspricht SMPPerSecond nicht genau 1s und
  //   SMPOfs wrde zeitlich wandern. Mit Hilfe von 'SMPOfsFrac' wird
  //   deswegen gelegentlich ein "Schalt-SMPOfs" eingefgt.
  s->SMPOfs++;
  s->SMPOfsFrac += PeriodsPerSample;
  if(s->SMPOfsFrac>=fCarrier)
  {
    s->SMPOfsFrac -= fCarrier;
    s->SMPOfs = 0;
    _State->SMPLoop++;
    // Position des ersten DCF77-PRNBit bestimmen
    s->DCF_SMPOfs_Next = PRNBitOfs_GetResult(&s->PRNBitOfs);
  }
  _State->SMPOfs = s->SMPOfs;
}

//################################################### TIM3_IRQHandler
void OPTS TIM3_IRQHandler()
{
  TDCF77Control* s = _Control;

  // Init Zeitpunkt fr nchsten IRQ
  {
    // auerdem wird der neue IRQ-Zeitpunkt um die Dauer einer DCF-Periode verrauscht:
    int PrevNoiseVal = ((s->Rand>>16)*ADCSamplesPerPeriod)>>16; // Zufallszahl 0..(ADCSamplesPerPeriod-1)
    s->Rand = s->Rand*214013+2531011;
    int NewNoiseVal = ((s->Rand>>16)*ADCSamplesPerPeriod)>>16; // Zufallszahl 0..(ADCSamplesPerPeriod-1)
    TIM3->CCR1 = TIM3->CCR1+AdcCntPerSample/IRQPerSample+NewNoiseVal-PrevNoiseVal;
    // durch das _bedingte_ Lschen wird auch bei 30ms lang blockiertem IRQ eine Eskalation verhindert:
    if((int16_t)(TIM3->CCR1-TIM3->CNT)>1) TIM3->SR=~TIM_SR_CC1IF;
  }

  // DCT von den neuen ADC-Samples durchfhren
  {
    DoDCT(s->TimedAdc.ValArr+s->AdcBufOfs, AdcCntPerSample/IRQPerSample/ADCSamplesPerPeriod, &s->CurC, &s->CurS);
    my_assert(PeriodsPerSample%IRQPerSample==0, _ErrAdcBufSize);
    // wenn folgendes my_assert() nur _vereinzelt_ fehlschlgt, dann drfte es entfernt werden
    my_assert((uint32_t)(AdcBufCnt-DMA1_Channel2->CNDTR-s->AdcBufOfs)>=AdcCntPerSample/IRQPerSample, _ErrAdcBufOverrun);
    s->AdcBufOfs += AdcCntPerSample/IRQPerSample;
    if(s->AdcBufOfs==AdcBufCnt) s->AdcBufOfs=0;
    s->IRQCur++;
  }

  // Sample verarbeiten
  if(s->IRQCur==IRQPerSample)
  {
    s->IRQCur = 0;
    IRQ_ProcessSample(s, s->CurC, s->CurS);
    // Debug: Samples als Streaming fr VisualStudio verfgbar machen (STM32->STLink->GDBServer->TCPIP->VisualStudio)
    if(DbgSample_Cnt!=0)
    {
      static int32_t DbgSample_Arr[DbgSample_Cnt];
      static int32_t DbgSample_Ofs = 0;
      DbgSample_Arr[(DbgSample_Ofs++)&(DbgSample_Cnt-1)] = s->CurC;
      DbgSample_Arr[(DbgSample_Ofs++)&(DbgSample_Cnt-1)] = s->CurS;
      TIM2->CCR3 = (uint32_t)DbgSample_Arr; // fr VisualStudio sichtbar machen (Missbrauch)
      TIM2->CCR4 = (uint32_t)DbgSample_Ofs; // fr VisualStudio sichtbar machen (Missbrauch)
    }
  }

  // Feinabgleich von fADC
  if((DMA1->ISR&DMA_ISR_TCIF2)!=0 && DMA1_Channel2->CNDTR<AdcBufCnt-1)
  {
    DMA1->IFCR = DMA_IFCR_CTCIF2;
    TimedAdc_Tweak(&s->TimedAdc);
  }
}

//################################################### DCF77_Init
void OPTS DCF77_Init(TDCF77Control* Control, TDCF77State* State, int32_t ppm)
{
  _Control = Control;
  _State   = State;

  for(int i=0; i<sizeof(*_Control); i++)
  {
    ((uint8_t*)_Control)[i] = 0;
  }
  for(int i=0; i<sizeof(*_State); i++)
  {
    ((uint8_t*)_State)[i] = 0;
  }

  PRNBitOfs_Reset(&Control->PRNBitOfs);

  // PA7: Analog
  RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
  GPIOA->CRL = (GPIOA->CRL&0x0fffffff)|0x00000000;

  // ADC_Frequenz einstellen (TimedAdc.TimArr[] definieren)
  int32_t* pSharedVal = TimedAdc_Init(&Control->TimedAdc, ppm);
  Carrier_Init(&Control->Carrier, pSharedVal);

  // TIM2
  // - fUpdate=fADC
  // - Trigger fr ADC-Sample-Start
  // - Trigger fr TIM3-Counter
  RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;
  TIM2->ARR = Control->TimedAdc.TimArr[0]; // Achtung: wird per DMA1_Channel2 manipuliert!
  TIM2->CCR1 = 10;                         // TIM3-Zhler-Trigger
  TIM2->CCR2 = 10;                         // ADC-Start-Trigger
  TIM2->CCER = TIM_CCER_CC2E;              // CC2 output enabled (fr ADC-Start-Trigger)
  TIM2->CCMR1 = TIM_CCMR1_OC2M_0*7;        // 7: PWM mode 2 - active TIMx_CNT>=TIMx_CCR1 (fr ADC-Start-Trigger)
  TIM2->CR2 = TIM_CR2_MMS_0*3;             // 3: Master Trigger Output "Compare CC1IF" (fr TIM3-Trigger)
  TIM2->DIER = TIM_DIER_UDE;               // Update Event (fr DMA1_Channel2-Trigger)

  // TIM3
  // - zhlt ADC-Samples
  // - hilft Strungen zu reduzieren (hilft Rauschen zu erzeugen)
  //   Beispiel:
  //   Bei 64MHz-STM32-Frequenz und STM32-Sleeps ("WFI()") gibts es
  //   deutliche Stromnderungen zwischen Sleep und der Bearbeitung in IRQ_ProcessSample().
  //   Da die Stromnderung immer zur gleichen Phase des Mischsignals
  //   (DoDCT8()) einsetzt, tritt als Folge bei Carrier() auch bei
  //   fADC/ADCSamplesPerPeriod ein "Phantom-Sender" auf. Ein schwaches
  //   DCF-Trgersignal kann dann nicht sicher erkannt werden. Mit Hilfe von TIM3
  //   kann der Zeitpunkt fr den IRQ und damit die Phase der Strung verndert
  //   werden.
  RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;
  TIM3->SMCR |= TIM_SMCR_TS_0*1;      // 1: Slave Trigger "Internal Trigger 1 (ITR1)" -> TIM2
  TIM3->SMCR |= TIM_SMCR_SMS_0*7;     // 7: Slave Trigger "Selected trigger clock the counter"
  TIM3->CCR1 = AdcCntPerSample/IRQPerSample; // allererster IRQ-Zeitpunkt
  TIM3->DIER = TIM_DIER_CC1IE;        // TIM3_IRQHandler
  TIM3->SR = ~TIM_SR_UIF;
  TIM3->CR1 = TIM_CR1_CEN;
  NVIC_EnableIRQ(TIM3_IRQn);

  // Init ADC
  my_assert(fMHz*1000*1000>32*fCarrier*ADCSamplesPerPeriod, _ErrAdcSampleTime); // ADC braucht mind. 32 CPU-Takte
  RCC->CFGR   |= RCC_CFGR_ADCPRE_0*0; // 0 -> PCLK2/2
  RCC->APB2ENR |= RCC_APB2ENR_ADC1EN;
  {
    // Calibrate ADC
    ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
    ADC1->CR2   |= ADC_CR2_ADON;        // start (single) conversion (1->1)
    while((ADC1->SR&ADC_SR_EOC)==0);    // wait for the end of conversion
    ADC1->CR2   |= ADC_CR2_RSTCAL;      // reset calibration
    while((ADC1->CR2&ADC_CR2_RSTCAL));  // wait for reset
    ADC1->CR2   |= ADC_CR2_CAL;         // start calibration
    while((ADC1->CR2&ADC_CR2_CAL)!=0);  // wait till calibration is done
    ADC1->CR2   &= ~ADC_CR2_ADON;       // ADC OFF
  }
  ADC1->CR2   |= ADC_CR2_DMA;         // DMA Enabled
  ADC1->CR2   |= ADC_CR2_EXTTRIG;     // External trigger
  ADC1->CR2   |= ADC_CR2_EXTSEL_0*3;  // 3 -> Trigger TIM2_CC2
  ADC1->SQR1  |= ADC_SQR1_L_0*(1-1);  // 0 -> 1 Kanal
  ADC1->SQR3  |= ADC_SQR3_SQ1_0*7;    // 7 -> PA7
  ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*0;  // 0 -> 14 Takte (T0+Tx=12.5+1.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*1;  // 1 -> 20 Takte (T0+Tx=12.5+7.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*2;  // 2 -> 26 Takte (T0+Tx=12.5+13.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*3;  // 3 -> 41 Takte (T0+Tx=12.5+28.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*4;  // 4 -> 54 Takte (T0+Tx=12.5+41.5) fr PA7
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC Start (1->1) (->1ExtraSample, damit TIM3 synchron zum Sample)

  // Init DMA1_CH1 zum Lesen der ADC-Ergebnisse
  RCC->AHBENR |= RCC_AHBENR_DMA1EN;
  DMA1_Channel1->CNDTR = AdcBufCnt;
  DMA1_Channel1->CPAR  = (uint32_t)&ADC1->DR;
  DMA1_Channel1->CMAR  = (uint32_t)Control->TimedAdc.ValArr;
  DMA1_Channel1->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *1) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *0) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // Init DMA1_CH2 zum Setzen der ADC-Samplezeitpunktabstnde
  DMA1_Channel2->CNDTR = AdcBufCnt;
  DMA1_Channel2->CPAR  = (uint32_t)&TIM2->ARR;
  DMA1_Channel2->CMAR  = (uint32_t)Control->TimedAdc.TimArr; // Zeitabstnde
  DMA1_Channel2->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *0) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *1) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // alles in Gang setzen
  TIM2->CR1 = TIM_CR1_CEN;
}

//################################################### DCF77_Deinit
void OPTS DCF77_Deinit(TDCF77Control* s)
{
  // Stop
  TIM2->CR1 = 0;
  // Reset TIM2
  RCC->APB1RSTR |= RCC_APB1RSTR_TIM2RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM2RST;
  RCC->APB1ENR &= ~RCC_APB1ENR_TIM2EN;
  // Reset TIM3
  RCC->APB1RSTR |= RCC_APB1RSTR_TIM3RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM3RST;
  RCC->APB1ENR &= ~RCC_APB1ENR_TIM3EN;
  // Reset ADC1
  RCC->APB2RSTR |= RCC_APB2RSTR_ADC1RST;
  RCC->APB2RSTR &= ~RCC_APB2RSTR_ADC1RST;
  RCC->APB2ENR &= ~RCC_APB2ENR_ADC1EN;
  // Reset DMA1_Channel2
  DMA1_Channel2->CCR = 0;
  // Reset DMA1_Channel1
  DMA1_Channel1->CCR = 0;
  // Reset PA7
  GPIOA->CRL = (GPIOA->CRL&0x0fffffff)|0x40000000;
}

//################################################### DCF77_Calib
void OPTS DCF77_Calib_fADC(TDCF77Control* s)
{
  bool InitTimedAdc = false;
  int ppm_L10 = Carrier_GetResult(&s->Carrier, &InitTimedAdc);
  if(InitTimedAdc)
  {
    s->OscHop = true;
    TimedAdc_Init(&s->TimedAdc, ppm_L10/1024);
  }
  // Fehler durch Drift bei zu kleinem Trgersignal verhindern
  if(PRNDriftFuse)
  {
    // Bei zu kleinem DCF77-Trgersignal driftet _DCF77State->SMPOfs. Ab einer
    // Drift von SamplesPerPRNBit/2 knnen falsche (scheinbar! gute)
    // Trefferwerte entstehen. Dieses passiert nur extrem selten.
    // Die Drift wird hier abgeschtzt und gegebenenfalls das Verwerfen
    // alter Daten erzwungen.
    enum
    {
      OSC_Startup_s       = 10,
      OSC_Threshold_dB_E1 = 10,
      dDrift0_E7          = 10,  // 1.0 ppm zustzliche Abweichung bei Drift
      dNoDrift_E7         =  5,  // je "PRNBitOfs_GetResult()" 0.5 ppm Korrektur
      nDCFBit             = 32700,   // int16_t-Wertebereich ausnutzen
      OSC_Const           = (int)(1e-7*nDCFBit*fCarrier/DCFPeriodsPerPRNBit*256),
    };
    int32_t dOSC_E7 = ppm_L10*10/1024;
    s->OSC_Time++;
    if(_State->Carrier_SNR_Fine_dB_E1<OSC_Threshold_dB_E1) s->OSC_Time=0;
    if(s->OSC_Time<OSC_Startup_s)
        {
          // noch nicht stabil
          if(_State->SMPLoop<OSC_Startup_s) s->OSC_Ref_e7=dOSC_E7;
          s->OSC_Drift += OSC_Const*(abs(s->OSC_Ref_e7-dOSC_E7)+dDrift0_E7)/256;
          if(s->OSC_Drift>nDCFBit/2) // halbes DCF-Bit
          {
            s->OscHop = true;
            s->OSC_Drift = 0;
          }
        }
      else
        {
          // stabil
          s->OSC_Time = OSC_Startup_s;
          s->OSC_Ref_e7 = dOSC_E7;
          if(s->OSC_Drift>0) s->OSC_Drift-=OSC_Const*dNoDrift_E7/256;
        }
  }
}

//################################################### DCF77_Suspend
void OPTS DCF77_Suspend(TDCF77Control* s)
{
  TIM2->CR1 = 0;
}

//################################################### DCF77_ResumeInit
void OPTS DCF77_ResumeInit(TDCF77Control* s)
{
  s->Carrier.FinePhaCnt = -1;  // -1 -> auch Raw-Wert der laufenden Messung verwerfen
  s->Carrier.AlwaysSmooth = true;
  DCTAvr512_Reset(&s->DCTAvr512);
}

//################################################### DCF77_Resume
void OPTS DCF77_Resume(TDCF77Control* s)
{
  TIM2->CR1 = TIM_CR1_CEN;
}
