
#ifndef DCF77DECBIT_H
#define DCF77DECBIT_H

#include "stdint.h"

//
// Einfacher _nicht_ fehlertoleranter Dekoder fr phasenmodulierte DCF77-Daten
// anhand von "BitVal"-Werten.
//
// Copyright: Public Domain
//
// C.B., 2021
//

typedef struct
{
  uint32_t ValidCnt;     // s.u. "ValidCnt"
  uint8_t  Zone_MEZS;    // 0..1; 0:Winterzeit
  uint8_t  Time_Second;  // 0..59
  uint8_t  Time_Minute;  // 0..59
  uint8_t  Time_Hour;    // 0..23
  uint8_t  Date_Day;     // 1..x
  uint8_t  Date_WeekDay; // 1..7; 1:Montag
  uint8_t  Date_Month;   // 1..12
  uint8_t  Date_Year;    // 0..99; 0:2000
  // intern
  uint32_t _DCFBits0;
  uint32_t _DCFBits1;
  uint32_t _DCFBits2;
  int32_t  _PrevBitCnt;
  int32_t  _BitCntOk;
} TDCF77DecBit;

//
// - "ValidCnt"
//   Gibt die Anzahl von aufeinanderfolgend zueinander passenden
//   (Prfbit- und Konsistenzprfung) DCF-Minutendatenstze an.
//   Der Wert spiegelt die Fehlerwahrscheinlichkeit wieder.
//   Zustzlich steigt die  Fehlerwahrscheinlichkeit mit der
//   Medauer. Die unteren Fehlerangaben beziehen sich auf eine
//   Mezeit von max. 10 Minuten.
//   - Gltigkeit der Werte in der TDCF77DecBit-Struktur in
//     Abhngigkeit von  "ValidCnt":
//     ValidCnt=0 : undefiniert
//     ValidCnt=1 : sehr fraglich; je nach Strungshufigkeit
//                  sind alle Fehlerwahrscheinlichkeiten mglich
//     ValidCnt=2 : ziemlich sicher; unter schlechten Bedingungen
//                  ist der Fehler etwa  1:100.000 (aus Simulation
//                  ermittelt)
//     ValidCnt>2 : uerst sicher
// - Zeitdifferenz / exakter Zeitbezug (innerhalb einer Sekunde)
//   Der nach Aufruf vom DCF77DecBit_ProcessVal() in TDCF77DecBit
//   angegebene Zeitstempel gilt fr die kommende Sekundenmarke.
//   Empfnger stellen Bits frhestens 995ms (s. Doku) nach einer
//   Sekundenmarke bereit. Die Dekoderzeit geht damit um etwa 5ms
//   vor.
//

void DCF77DecBit_Init(TDCF77DecBit* s);

void DCF77DecBit_Deinit(TDCF77DecBit* s);

void DCF77DecBit_ProcessVal(TDCF77DecBit* s,
                            int32_t DCF77_BitCnt,
                            int32_t DCF77_BitVal);
  // - Parameter
  //   - s            : Struktur sollte vor ersten Aufruf mit
  //                    DCF77DecBit_Init() initialisiert werden
  //   - DCF77_BitCnt : laufender Zhler fr das DCF-Bit;
  //                    dient zur Erkennung einer Sequenz; eine nderung
  //                    ungleich 1 zum vorherigen Aufruf bewirkt einen
  //                    Reset der Dekodierung
  //   - DCF77_BitVal : 0..1; das DCF-Bit

#endif
