
#include "stm32f10x.h"
#include "stdlib.h"
#include "stdbool.h"
#include "stdarg.h"

#include "DCF77.h"
#include "DCF77DecHit.h"
#include "DCF77DecBit.h"

//################################################### Set_HSE
static int Set_HSE(int MHz)
{
  // (Einweg - nicht wiederverwenden)

  // 1, 2, 4, 8
  //  8.. 64 in 4MHz Schritten
  // 64..136 in 8MHz Schritten

  // (schmutziger) Reset
  RCC->CFGR = 0;
  RCC->CR &= ~(RCC_CR_HSEON|RCC_CR_PLLON);
  FLASH->ACR &= ~FLASH_ACR_LATENCY;

  // HSE einschalten
  RCC->CR |= RCC_CR_HSEON;

  // warten bis HSE luft
  while((RCC->CR|RCC_CR_HSERDY)==0)
  {
  }

  // warten bis HSE Systemclock ist
  RCC->CFGR |= RCC_CFGR_SW_0*1;
  while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*1)
  {
  }

  if(MHz<=8)
  {
    if(MHz==- 16) RCC->CFGR |= RCC_CFGR_HPRE_DIV512; // kein STLINK Zugriff
    if(MHz==- 32) RCC->CFGR |= RCC_CFGR_HPRE_DIV256;
    if(MHz==- 64) RCC->CFGR |= RCC_CFGR_HPRE_DIV128;
    if(MHz==-125) RCC->CFGR |= RCC_CFGR_HPRE_DIV64;
    if(MHz==-500) RCC->CFGR |= RCC_CFGR_HPRE_DIV16;
    if(MHz==1) RCC->CFGR |= RCC_CFGR_HPRE_DIV8;
    if(MHz==2) RCC->CFGR |= RCC_CFGR_HPRE_DIV4;
    if(MHz==4) RCC->CFGR |= RCC_CFGR_HPRE_DIV2;
    if(MHz==8) RCC->CFGR |= RCC_CFGR_HPRE_DIV1;
    return 0;
  }

  // PLL auf HSE stellen
  RCC->CFGR |= RCC_CFGR_PLLSRC;

  // PLL braucht bei 8MHz ca. 0.1mA zu
  if(MHz>24)
  {
    // WaitStates
    if(MHz>48)
        FLASH->ACR |= FLASH_ACR_LATENCY_1*2;
      else
        FLASH->ACR |= FLASH_ACR_LATENCY_1*1;
  }
  if((MHz&4)!=0)
  {
    RCC->CFGR |= RCC_CFGR_PLLXTPRE;
    MHz *= 2;
  }
  RCC->CFGR = (RCC->CFGR&0xffc3ffff)|(((MHz>>3)-2)<<18);
  RCC->CR |= RCC_CR_PLLON;
  while((RCC->CR&RCC_CR_PLLRDY)==0)
  {
  }
  RCC->CFGR = (RCC->CFGR&~RCC_CFGR_SW)|(RCC_CFGR_SW_0*2);
  while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*2)
  {
  }
  return 0;
}

//################################################### uprintf
static void uprintf_chrout(char Char);
static int  uprintf_GetNum(const char** Fmt);
static void uprintf(const char* Fmt, ...)
{
  // "mini-usart-printf"
  // (Einweg - nicht wiederverwenden)
  va_list va;
  va_start(va, Fmt);
  for(; *Fmt!=0; Fmt++)
  {
    if(*Fmt=='%')
      {
        bool Zero = *(++Fmt)=='0';
        int  n0 = uprintf_GetNum(&Fmt);
        int  n1 = Zero ? n0 : 0;
        if(Zero) n1=n0;
        if(*Fmt=='.')
        {
          Fmt++;
          n1 = uprintf_GetNum(&Fmt)+1;
        }
        switch(*Fmt)
        {
          case 'c': {
                      uprintf_chrout(va_arg(va, int32_t));
                      break;
                    }
          /*case 's': {
                      const char* s = va_arg(va, const char*);
                      if(n0!=0)
                      {
                        int n2 = 0;
                        for(; s[n2]!=0; n2++) { };
                        for(; n2<n0; n2++) uprintf_chrout(' ');
                      }
                      while(*s!=0) uprintf_chrout(*s++);
                      break;
                    }*/
          case 'i': {
                      int v = va_arg(va, int32_t);
                      int cv = 10;
                      int n2 = 1;
                      for(; abs(v)>=cv; n2++) cv*=10;
                      int nn = n1>n2 ? n1 : n2;
                      int nSpace = n0-nn;
                      if(v<0) nSpace--;
                      for(; nSpace>0; nSpace--) uprintf_chrout(' ');
                      if(v<0)
                      {
                        uprintf_chrout('-');
                        v = abs(v);
                      }
                      for(int i=0; i<nn; i++)
                      {
                        if(i>=n1-n2) cv/=10;
                        int n = v/cv;
                        uprintf_chrout(n+48);
                        v -= n*cv;
                        if(!Zero && i>=n2-n1)
                        {
                          uprintf_chrout('.');
                          Zero = true;
                        }

                      }
                      break;
                    }
        }
      }
    else
      uprintf_chrout(*Fmt);
  }
  va_end(va);
}
static void uprintf_chrout(char Char)
{
  // TXD-Ausgang ist GPIO_PA9
  const int Baud = 9600;
  if((USART1->CR1&USART_CR1_UE)==0)
  {
    RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
    GPIOA->CRH = (GPIOA->CRH&0xffffff0f)|0x000000a0;
    RCC->APB2ENR |= RCC_APB2ENR_USART1EN;
    USART1->CR1 = USART_CR1_UE;
    int divval = (fMHz/1*1000000+Baud/2)/Baud; // 16..65535
    USART1->BRR = divval;
    USART1->CR1 |= USART_CR1_TE;
  }
  USART1->DR = Char;
  while((USART1->SR&USART_SR_TXE)==0)
  {
    __WFI();
  }
}
static int uprintf_GetNum(const char** Fmt)
{
  int v = 0;
  while(**Fmt>='0' && **Fmt<='9')
  {
    v = v*10+(*(*Fmt)++-'0');
  }
  return v;
}

//################################################### main0_USART_LCResonanceMeter
int main_LCResonanceMeter(void);
int main0_USART_LCResonanceMeter(void)
{
  // Hilfsprogramm um die Resonanzfrequenz eines LC-Schwingkreises zu
  // messen. Weitere Infos (wie Hardware-Anschluss) stehen in der Datei
  // "main_LCResonanceMeter.c".
  return main_LCResonanceMeter();
}

//################################################### main1_LED_SignalQuality
int main1_LED_SignalQuality(void)
{
  //
  // Empfangsqualitt durch Blinken der Bluepill-LED anzeigen
  //
  //   0 x blinken  unbrauchbar (Trefferzahl   0..19 Bits)
  //   1 x blinken  schlecht    (Trefferzahl  20..40 Bits)
  //   2 x blinken  gut         (Trefferzahl  40..60 Bits)
  //   3 x blinken  sehr gut    (Trefferzahl 60..256 Bits)
  //

  // GPIO_PC13 (LED) auf Ausgang (50MHz)
  RCC->APB2ENR |= RCC_APB2ENR_IOPCEN;
  GPIOC->CRH = (GPIOC->CRH&0xff0fffff)|0x00300000;
  GPIOC->BSRR = 1<<(13+0*16); // LED aus

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);

  int MaxHitVal = 0;
  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    if(abs(DCF77State.DCF77_HitVal)>MaxHitVal)
    {
      MaxHitVal = abs(DCF77State.DCF77_HitVal);
    }
    // alle zwei Sekunden blinken lassen
    if((TotalBitCnt&1)!=0)
    {
      int nBlink = 0;
      if(MaxHitVal>=20) nBlink=1;
      if(MaxHitVal>=40) nBlink=2;
      if(MaxHitVal>=60) nBlink=3;
      MaxHitVal = 0;
      for(int t=0; t<nBlink; t++)
      {
        GPIOC->BSRR = 1<<(13+1*16); // LED an
        for(int i=0; i<1200; i++) __WFI(); // etwa 250ms warten
        GPIOC->BSRR = 1<<(13+0*16); // LED aus
        for(int i=0; i<1200; i++) __WFI(); // etwa 250ms warten
      }
    }
  }
  return 0;
}

//################################################### main2_LED_MinuteSignal
int main2_LED_MinuteSignal(void)
{
  //
  // Bluepill-LED fr die erste Sekunde einer Minute einschalten.
  // Solange unbekannt, Dauerleuchten.
  //

  // GPIO_PC13 (LED) auf Ausgang (50MHz)
  RCC->APB2ENR |= RCC_APB2ENR_IOPCEN;
  GPIOC->CRH = (GPIOC->CRH&0xff0fffff)|0x00300000;
  GPIOC->BSRR = 1<<(13+0*16); // LED aus

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;
  TDCF77DecHit  DCF77Dec;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);
  DCF77DecHit_Init(&DCF77Dec);

  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    DCF77DecHit_ProcessVal(&DCF77Dec, DCF77State.DCF77_BitCnt, DCF77State.DCF77_HitVal);
    if(DCF77Dec.Time_dbErr<60 || DCF77Dec.Time_Second==0)
        GPIOC->BSRR = 1<<(13+1*16); // LED an
      else
        GPIOC->BSRR = 1<<(13+0*16); // LED aus
  }
  return 0;
}

//################################################### GenerateSecondSyncOnPin_GPIO_PB0
static void GenerateSecondSyncOnPin_GPIO_PB0(TDCF77State* s)
{
  // Sekundenimpuls an GPIO_PB0 erzeugen
  //
  // - bei DCF77-Sekundenbeginn steigende Flanke an PB0
  // - die Abweichung ist besser als +-0.7ms
  // - die Impulsdauer ist mind. 70ms
  // - bei schlechtem Signal wird kein Impuls ausgegeben
  // - Aufruf von GenerateSecondSyncOnPin_GPIO_PB0()
  //   - mindestens ein Aufruf  70ms bis   1ms vor  Sekundenbeginn
  //   - mindestens ein Aufruf 300ms bis 900ms nach Sekundenbeginn
  // - SamplesPerPRNBit muss 6 sein, da der "Schalt-SMPOfs" (s. "DCF77.c")
  //   in der Berechnung von "SecImpulsTimerValue" fehlt)
  //
  // Es wre nur kaum ungenauer den Sekundenimpuls mit SMPOfs, DCF77_SMPOfs
  // und GPIO_BSRR als "Dreizeiler" umzusetzen. Dies Beispiel soll aber den
  // Zusammenhang zwischen DCF77-, TIM3- und der TDCF77State-Zeit
  // "SMPLoop & SMPOfs" zeigen.

  const int MinAvrAbsHitVal = 20; // Schwellwert - um fragliche Impulse zu auszulassen

  // Init Hardware
  int CurMode = TIM3->CCMR2&TIM_CCMR2_OC3M;
  if(CurMode==0)
  {
    // PB0: Alternate function output Push-pull / 2MHz
    RCC->APB2ENR |= RCC_APB2ENR_IOPBEN;
    GPIOB->CRL = (GPIOB->CRL&0xfffffff0)|0x0000000a;
    TIM3->CCER |= TIM_CCER_CC3E;
  }
  // TIM3 fr Impulsbeginn initialisieren
  int n = s->SMPLoop;    // #1
  int p = s->SMPOfs;     // #2
  if(n!=s->SMPLoop) p=0; // #3 falls berlauf whrend (IRQ) letzten 3 Zeilen
  int dp = SMPPerSecond+s->DCF77_SMPOfs-p;
  while(dp>+SMPPerSecond/2) dp-=SMPPerSecond;
  int dTimTicks = dp*(ADCSamplesPerPeriod*DCFPeriodsPerPRNBit/SamplesPerPRNBit);
  if(dp>SamplesPerPRNBit && dTimTicks<0xffff)
  {
    // Der Sekundenbeginn liegt in den nchsten 65536 TIM3-Takten.
    // (der 16-Bit-Timer kann Sekundenbeginn also behandeln)
    if(s->DCF77_AvrAbsHitVal>=MinAvrAbsHitVal && CurMode==TIM_CCMR2_OC3M_0*2)
    {
      uint16_t SecImpulsTimerValue = (n*SMPPerSecond+p)*(ADCSamplesPerPeriod*DCFPeriodsPerPRNBit/SamplesPerPRNBit)+dTimTicks;
      // SecImpulsTimerValue hat nun den TIM3->CNT-Sollwert vom nchsten Sekundenbeginn
      TIM3->CCR3 = SecImpulsTimerValue;
      TIM3->CCMR2 = (TIM3->CCMR2&~TIM_CCMR2_OC3M)|(TIM_CCMR2_OC3M_0*1); // 1 : Set channel 3 to active level on match
      TIM3->SR = ~TIM_SR_CC3IF;
    }
  }
  // TIM3 fr Impulsende initialisieren (eine TIM3-Periodendauer)
  if((TIM3->SR&TIM_SR_CC3IF)!=0)
  {
    TIM3->CCMR2 = (TIM3->CCMR2&~TIM_CCMR2_OC3M)|(TIM_CCMR2_OC3M_0*2); // 2 : Set channel 3 to inactive level on match
  }
}

//################################################### GenerateFrequencyStandardOnPin_GPIO_PB6
void GenerateFrequencyStandardOnPin_GPIO_PB6(TDCF77State* s)
{
  // Eine zur DCF77-Trgerfrequenz synchrone Frequenz an GPIO_PB6 erzeugen.
  // (siehe auch "GenerateFrequencyStandardOnPin_GPIO_PA8")
  //
  // - die Phasenlage ist (einmalig) zufllig
  // - die Phasenlage ist dauerhaft ("Atomzeit") bis auf +-3s konstant
  //   (zuzglich sonstigen Fehlern (+-10s?; s. "PTB-Mitteilungen 114 (2004), Heft 4")
  // - das Signal wird erst bei ausreichend sicherem Empfang eingeschaltet
  //   (mind. 10s)
  // - das Signal wird bei schlechtem Empfang _dauerhaft_ abgeschaltet
  //   (STM32-Reset erforderlich)
  //   -> damit ist sichergestellt, dass ein vorhandenes GPIO_PB6-Signal
  //      immer synchron ist
  // - Taktverhltnis 1:1
  // - die Frequenz kann in in Grenzen vorgegeben werden
  //   - der Faktor 930000Hz/fSollHz muss ganzzahlig sein
  //   - der Faktor 930000Hz/fSollHz muss zwischen ca. 10 und 65500 liegen
  //   - fCarrier sollte kein ungradzahliges Vielfaches von fSollHz (z.B. 77500Hz,
  //     15500Hz sein, da die Zuleitung als Sendeantenne fr einen DCF77-Scheinsender
  //     wirkt
  // - der Aufruf von GenerateFrquencyStandardOnPin_GPIO_PB6() muss immer
  //   nach dem Aufruf von DCF77_Calib_fADC() erfolgen
  //
  // - Empfehlungen
  //   - Antennenspule gut ausrichten (dPhi<+-30)
  //   - stabiles Signal (DCF77_AvrAbsHitVal>30)
  //   - ADCSamplesPerPeriod==12 (Maximalwert -> minimaler Jitter)

  enum
  {
     fSollHz = 20, // z.B. 20Hz, 50Hz, 100Hz, 200Hz, 500Hz, 1000Hz
     StartWarmupUpTime_s  = 10, // "TCarrier" braucht das
     StartMinAvrAbsHitVal = 20, // um echten (statt Phantom-Sender) zu erkennen
     StopMinAvrAbsHitVal  = 15, //
     StopMinFine_dB_E1    = 20, // "DCF77-Umpolung" (durch kurzeitigem
     //                         // Rckgang) erkennen (Umpolung z.B. durch
                                // Drehen der Spule um 180)
     // berechnet
     TimCyclesPerPeriod = fCarrier*ADCSamplesPerPeriod/fSollHz,
  };

  static uint16_t _DMAArr[2];
  static int32_t  _dARR_L8;
  static int32_t  _RstCnt;

  // Init Hardware
  if(TIM4->CR1==0)
  {
    // "fSollHz" prfen
    while(fSollHz*TimCyclesPerPeriod!=fCarrier*ADCSamplesPerPeriod)
    {
      // Fehler: fSollHz passt nicht
    }
    // "ADCSamplesPerPeriod" prfen
    while(ADCSamplesPerPeriod!=12)
    {
      // Fehler: Jitter ist nicht optimal (ADCSamplesPerPeriod ist nicht maximal)
    }
    // die Phasenabweichung fADC zu DCF-Trger muss verfgbar sein
    while(CarrierDoCalc_dCycles==0)
    {
      // Fehler: CarrierDoCalc_dCycles nicht aktiv
    }
    // Prinzip
    // Mit TIM4 wird das PWM-Signal erzeugt. Die Taktquelle fr TIM4 ist der
    // ADC-Takt (TIM2_SR_UIF). Aus "TDCF77State.Carrier_dCycles_L8" kann die
    // Abweichung der ADC-Takte berechnet werden. TIM4->ARR wird gegebenenfalls
    // angepasst.

    // PB6: Alternate function output Push-pull / 2MHz
    RCC->APB2ENR |= RCC_APB2ENR_IOPBEN;
    GPIOB->CRL = (GPIOB->CRL&0xf0ffffff)|0x0a000000;
    // TIM4-Takt vom ADC-Wandler bernehmen
    RCC->APB1ENR |= RCC_APB1ENR_TIM4EN;
    TIM4->ARR = TimCyclesPerPeriod-1;
    TIM4->CCR1 = TimCyclesPerPeriod/2;
    TIM4->SMCR |= TIM_SMCR_TS_0*1;      // 1: Slave Trigger "Internal Trigger 1 (ITR1)" -> TIM2
    TIM4->SMCR |= TIM_SMCR_SMS_0*7;     // 7: Slave Trigger "Selected trigger clock the counter"
    TIM4->CCMR1 |= TIM_CCMR1_OC1M_0*7;  // 7: PWM mode 2 - inactive as long as TIMx_CNT<TIMx_CCR1
    TIM4->CCER |= TIM_CCER_CC1E;
    TIM4->DIER = TIM_DIER_UDE;          // Update Event (fr DMA1_Channel7-Trigger)
  }

  // Start
  if(s->DCF77_BitCnt>=StartWarmupUpTime_s && s->DCF77_AvrAbsHitVal>=StartMinAvrAbsHitVal && TIM4->CR1==0)
  {
    _RstCnt = s->DCF77_RstCnt;
    TIM4->CR1 = TIM_CR1_CEN;
    _dARR_L8 = s->Carrier_dCycles_L8*ADCSamplesPerPeriod;
  }

  // Stop
  if(s->DCF77_RstCnt!=_RstCnt || s->DCF77_AvrAbsHitVal<StopMinAvrAbsHitVal || s->Carrier_SNR_Fine_dB_E1<StopMinFine_dB_E1)
  {
    if(TIM4->CR1!=0) TIM4->CCER=0;
  }

  // bei groer Abweichung einen PWM-Takt (ca. 1.1s) einfgen bzw. entfernen
  int ddArr_L8 = s->Carrier_dCycles_L8*ADCSamplesPerPeriod-_dARR_L8;
  if((ddArr_L8<-128 || ddArr_L8>128) && TIM4->CR1!=0 && TIM4->CCER!=0)
  {
    // DMA verwenden um anderes ARR genau einmal anzuwenden (Pfusch)
    int d = ddArr_L8>0 ? +1 : -1;
    _dARR_L8 += d*256;
    _DMAArr[0] = TimCyclesPerPeriod-1-d; // einmaliger Wert
    _DMAArr[1] = TimCyclesPerPeriod-1-0; // normaler Wert
    DMA1_Channel7->CCR   = 0;
    DMA1_Channel7->CNDTR = 2;
    DMA1_Channel7->CPAR  = (uint32_t)&TIM4->ARR;
    DMA1_Channel7->CMAR  = (uint32_t)_DMAArr;
    DMA1_Channel7->CCR = (DMA_CCR1_MEM2MEM *0) |
                         (DMA_CCR1_PL_0    *0) |
                         (DMA_CCR1_MSIZE_0 *1) |
                         (DMA_CCR1_PSIZE_0 *1) |
                         (DMA_CCR1_MINC    *1) |
                         (DMA_CCR1_PINC    *0) |
                         (DMA_CCR1_CIRC    *0) |
                         (DMA_CCR1_DIR     *1) |
                         (DMA_CCR1_TEIE    *0) |
                         (DMA_CCR1_HTIE    *0) |
                         (DMA_CCR1_TCIE    *0) |
                         (DMA_CCR1_EN      *1);
  }
}

//################################################### GenerateFrequencyStandardOnPin_GPIO_PA8
static void GenerateFrequencyStandardOnPin_GPIO_PA8(TDCF77State* s)
{
  // Eine zur DCF77-Trgerfrequenz synchrone Frequenz an GPIO_PA8 erzeugen.
  //
  // - die Phasenlage ist (einmalig) zufllig
  // - die Phasenlage ist dauerhaft ("Atomzeit") bei gutem Empfang bis auf +-1s
  //   konstant (zuzglich sonstigen Fehlern (+-10s?; s. "PTB-Mitteilungen 114
  //   (2004), Heft 4")
  // - das Signal wird erst bei ausreichend sicherem Empfang eingeschaltet
  //   (mind. 10s)
  // - das GPIO_A8-Signal wird bei schlechtem Empfang _dauerhaft_ abgeschaltet
  //   (STM32-Reset erforderlich)
  //   -> damit ist sichergestellt, dass ein vorhandenes GPIO_PA8-Signal
  //      immer synchron ist
  // - Taktverhltnis 1:1
  // - die Frequenz kann in in Grenzen vorgegeben werden
  //   - fSollHz muss ganzzahliger Teiler von 465000Hz sein
  //   - fSollHz muss grer als ein 128000tel von fCPU sein
  // - der Aufruf von GenerateFrquencyStandardOnPin_GPIO_PA8() muss immer
  //   nach dem Aufruf von DCF77_Calib_fADC() erfolgen
  //
  // - Empfehlungen
  //   - Antennenspule gut ausrichten (dPhi<+-20)
  //   - stabiles und starkes Signal (DCF77_AvrAbsHitVal>60)
  //   - strungsarme Umgebung
  //   - Umgebungstemperatur solle sich nur sehr langsam ndern
  //   - Temperaturausgleich abwarten (Carrier_dCycles_L8 fr lngere
  //     Zeit <+-25; Temperaturverhalten kann durch CarrierFineSeconds=2
  //     erheblich beschleunigt werden)
  //   - ADCSamplesPerPeriod==12

  enum
  {
     fSollHz = 1000, // z.B. 250Hz, 500Hz, 1000Hz, 2500Hz, 5000Hz
     StartMinAvrAbsHitVal = 30, // um echten (statt Phantom-Sender) zu erkennen
     StopMinAvrAbsHitVal  = 20, //
     TempGradCorr_s       = 1*CarrierFineSeconds, // Temperaturnderung vorausberechnen
     // berechnet
     TimCyclesPerPeriod = fCarrier*ADCSamplesPerPeriod/2/fSollHz,
  };

  static int32_t _DCFLocked;
  static int16_t _Arr[TempGradCorr_s];
  static int16_t _Ofs;

  // Init Hardware
  if(TIM4->CR1==0)
  {
    // "fSollHz" prfen
    while(fSollHz*TimCyclesPerPeriod!=fCarrier*ADCSamplesPerPeriod/2)
    {
      // Fehler: fSollHz passt nicht
    }
    // "ADCSamplesPerPeriod" prfen
    while(ADCSamplesPerPeriod!=12)
    {
      // Fehler: Jitter ist nicht optimal (ADCSamplesPerPeriod ist nicht maximal)
    }
    // fADC muss mit DCF-Trger synchron gehalten werden
    while(CarrierDoSynchronize==0)
    {
      // Fehler: CarrierDoSynchronize-Option ist nicht eingeschaltet
    }
    // Prinzip
    // Mit TIM4 wird der ADC-Sampletakt (TIM2) geteilt. Das Ergebnis wird mit
    // TIM1 (entsprechend der aktuellen Abweichung Carrier_dCycles_L8)
    // zeitlich verzgert und als PWM-Signal ausgegeben.

    // PA8: Alternate function output Push-pull / 50MHz
    RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
    GPIOA->CRH = (GPIOA->CRH&0xfffffff0)|0x0000000b;
    // TIM4: Takt als Slave vom TIM2 bernehmen, teilen und als Master bereitstellen
    RCC->APB1ENR |= RCC_APB1ENR_TIM4EN;
    TIM4->ARR = TimCyclesPerPeriod-1;
    TIM4->SMCR |= TIM_SMCR_TS_0*1;      // 1: Slave Trigger "Internal Trigger 1 (ITR1)" -> TIM2
    TIM4->SMCR |= TIM_SMCR_SMS_0*7;     // 7: Slave Trigger "Selected trigger clock the counter"
    TIM4->CR2 |= TIM_CR2_MMS_0*2;       // 2: Master Trigger Output "The update event is selected as trigger output (TRGO)"
    // TIM1 fr die Zeitverzgerung
    RCC->APB2ENR |= RCC_APB2ENR_TIM1EN;
    TIM1->BDTR |= TIM_BDTR_MOE;
    TIM1->SMCR |= TIM_SMCR_TS_0*3;      // 3: Slave Trigger "Internal Trigger 3 (ITR3)" -> TIM4
    TIM1->SMCR |= TIM_SMCR_SMS_0*4;     // 4: Reset Mode - Rising edge of the selected trigger input (TRGI) reinitializes the counter.
    TIM1->CCMR1 |= TIM_CCMR1_OC1M_0*3;  // 3: Toggle - OC1REF toggles when TIMx_CNT=TIMx_CCR1
    TIM1->CCMR1 |= TIM_CCMR1_OC1PE;     // Output Compare 1 preload enable
    TIM1->CCER |= TIM_CCER_CC1E;        // Capture/Compare 1 output enable
    TIM1->CCR2 = 0xffff;
  }
  // Start
  if(s->Carrier_Synchronized!=0 && s->DCF77_AvrAbsHitVal>=StartMinAvrAbsHitVal && TIM4->CR1==0)
  {
    _DCFLocked = s->Carrier_Synchronized;
    for(int i=0; i<TempGradCorr_s; i++)
    {
      _Arr[i] = s->Carrier_dCycles_L8;
    }
    TIM4->CR1 = TIM_CR1_CEN;
    TIM1->CR1 = TIM_CR1_CEN;
  }
  // Stop
  if(_DCFLocked!=s->Carrier_Synchronized || s->DCF77_AvrAbsHitVal<StopMinAvrAbsHitVal || (TIM1->SR&TIM_SR_CC2IF)!=0)
  {
    if(TIM1->CR1!=0) TIM1->CCER=0;
  }
  // Zeitverzgerung setzen
  int dDCF_Cycles_L8 = s->Carrier_dCycles_L8; // "alt"
  // - bei Temperaturnderungen hinkt Carrier_dCycles_L8 um 5s (=CarrierFineSeconds/2) hinterher
  // - hier wird versucht (besser wre ein kleineres CarrierFineSeconds) den aktuellen Wert zu schtzen
  if(TempGradCorr_s>0)
  {
    int nCorr = s->Carrier_dCycles_L8-_Arr[_Ofs];
    _Arr[_Ofs++] = s->Carrier_dCycles_L8;
    if(_Ofs==TempGradCorr_s) _Ofs=0;
    dDCF_Cycles_L8 += nCorr*CarrierFineSeconds/(TempGradCorr_s*2);
  }
  int n = fMHz*1000000/4/fSollHz + dDCF_Cycles_L8*(fMHz*1000000/fCarrier)/256;
  if(n<5 || n>90*(fMHz*1000000/fSollHz)/100-5 || n>65000) TIM1->CCR2=0; // nicht ausgleichbar -> Stop erzwingen
  TIM1->CCR1 = n;
}

//################################################### main3_GPIO_SyncSignals
int main3_GPIO_SyncSignals(void)
{
  // GPIO_PB0: Sekundenimpuls
  //           s.o. "GenerateSecondSyncOnPin_GPIO_PB0()"
  // GPIO_PA8: Frequenznormal
  //           s.o. "GenerateFrequencyStandardOnPin_GPIO_PA8()"

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);

  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
      GenerateSecondSyncOnPin_GPIO_PB0(&DCF77State);
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    GenerateFrequencyStandardOnPin_GPIO_PA8(&DCF77State);
  }
  return 0;
}

//################################################### main4_USART_DCF77
int main4_USART_DCF77(void)
{
  // TDCF77-Moduldaten an USART1 (9600/8-N-1 an GPIO_PA9) ausgeben

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);

  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    uprintf("%2i Err |",
            DCF77State.Error);
    uprintf(" %5i Loop %4i Ofs  |",
            DCF77State.SMPLoop,
            DCF77State.SMPOfs);
    uprintf(" %3i dB  %3.1i dB %6.2i Cycles %4i Locked %5.3i ppm |",
            DCF77State.Carrier_SNR_Raw_dB,
            DCF77State.Carrier_SNR_Fine_dB_E1,
            100*DCF77State.Carrier_dCycles_L8/256,
            DCF77State.Carrier_Synchronized,
            DCF77State.Carrier_dfHSE_E9);
    uprintf("%5i uV |",
            DCF77State.Signal_Ueff_uV);
    uprintf(" %4i Ofs  %3i Rst  %3i Avr  %4i Hit  %c Bit  %5i Cnt  %5i Total\r\n",
            DCF77State.DCF77_SMPOfs,
            DCF77State.DCF77_RstCnt,
            DCF77State.DCF77_AvrAbsHitVal,
            DCF77State.DCF77_HitVal,
            DCF77State.DCF77_BitVal ? 'X' : '.',
            DCF77State.DCF77_BitCnt,
            DCF77State.DCF77_TotalBitCnt);
  }
  return 0;
}

//################################################### main5_USART_DecodeBitValue
int main5_USART_DecodeBitValue(void)
{
  // Die DCF-Bitwerte vom TDCF77-Modul vom TDCF77-Modul per TDCF77DecBit-Modul
  // dekodieren und Ergebnis an USART1 (9600/8-N-1 an GPIO_PA9) ausgeben.

  static const char* WeekdayStr = "--MoDiMiDoFrSaSo";

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;
  TDCF77DecBit  DCF77Dec;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);
  DCF77DecBit_Init(&DCF77Dec);

  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    DCF77DecBit_ProcessVal(&DCF77Dec, DCF77State.DCF77_BitCnt, DCF77State.DCF77_BitVal);
    uprintf("%5i TotalBitCnt  %5i BitCnt  %3i AvrAbsHitVal  |",
            DCF77State.DCF77_TotalBitCnt,
            DCF77State.DCF77_BitCnt,
            DCF77State.DCF77_AvrAbsHitVal);
    uprintf("%5i uV |",
            DCF77State.Signal_Ueff_uV);
    uprintf("  %c%c,  %02i.%02i.20%02i  %02i:%02i:%02i MEZ%c - %3i ValidCnt\r\n",
            WeekdayStr[2*DCF77Dec.Date_WeekDay+0],
            WeekdayStr[2*DCF77Dec.Date_WeekDay+1],
            DCF77Dec.Date_Day,
            DCF77Dec.Date_Month,
            DCF77Dec.Date_Year,
            DCF77Dec.Time_Hour,
            DCF77Dec.Time_Minute,
            DCF77Dec.Time_Second,
            DCF77Dec.Zone_MEZS!=0 ? 'S' : ' ',
            DCF77Dec.ValidCnt);
  }
  return 0;
}

//################################################### main6_USART_DecodeHitValue
int main6_USART_DecodeHitValue(void)
{
  // Die Trefferwerte fr die PRN-Kodes (Pseudo-Random Noise Kode) vom
  // TDCF77-Modul per TDCF77DecHit-Modul dekodieren und Ergebnis an USART1
  // (9600/8-N-1 an GPIO_PA9) ausgeben.

  static const char* WeekdayStr = "--MoDiMiDoFrSaSo";

  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;
  TDCF77DecHit  DCF77Dec;

  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);
  DCF77DecHit_Init(&DCF77Dec);

  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    DCF77DecHit_ProcessVal(&DCF77Dec, DCF77State.DCF77_BitCnt, DCF77State.DCF77_HitVal);
    uprintf("%5i TotalBitCnt  %5i BitCnt  |",
            DCF77State.DCF77_TotalBitCnt,
            DCF77State.DCF77_BitCnt);
    uprintf("%5i uV |",
            DCF77State.Signal_Ueff_uV);
    uprintf(" %3i AvrHitVal  %c%c,  %02i.%02i.20%02i  %02i:%02i:%02i  MEZ%c   dbErr[%3i Date, %3i Time  %3i Zone]\r\n",
            DCF77Dec.AvrHitVal,
            WeekdayStr[2*DCF77Dec.Date_WeekDay+0],
            WeekdayStr[2*DCF77Dec.Date_WeekDay+1],
            DCF77Dec.Date_Day,
            DCF77Dec.Date_Month,
            DCF77Dec.Date_Year,
            DCF77Dec.Time_Hour,
            DCF77Dec.Time_Minute,
            DCF77Dec.Time_Second,
            DCF77Dec.Zone_MEZS!=0 ? 'S' : ' ',
            DCF77Dec.Date_dbErr,
            DCF77Dec.Zone_dbErr,
            DCF77Dec.Time_dbErr);
  }
  return 0;
}

//################################################### main7_SuspendResume
int main7_SuspendResume(void)
{
  // Verarbeitung im TDCF77-Modul alle 7 Sekunden fr eine Sekunde
  // aussetzen.
  // Die variable Initialisierungszeit kann damit vorverlagert werden. Der
  // STM32 kann so im Tiefschlaf (z.B. 100 Sekunden) stromsparend bis einige
  // Millisekunden vor dem Senden eines bestimmten DCF-Bits warten um dann
  // den (stromintensiven) Empfang fr z.B. 1 Bit wieder zu aktivieren.
  TDCF77State   DCF77State;
  TDCF77Control DCF77Control;
  Set_HSE(fMHz);
  DCF77_Init(&DCF77Control, &DCF77State, 0);
  for(int TotalBitCnt=0; ; TotalBitCnt++)
  {
    while(TotalBitCnt==DCF77State.DCF77_TotalBitCnt)
    {
      __WFI();
    }
    DCF77_Calib_fADC(&DCF77Control); // fADC-Kalibrierung nicht vergessen!
    if((DCF77State.SMPLoop&7)==7)
    {
      int sSleep = 3;
      int32_t SleepTime = sSleep*fMHz*1000*1000;
      SleepTime += sSleep*DCF77State.Carrier_dfHSE_E9*fMHz/1000;
      DWT->CTRL = 0x40000001;
      DCF77_Suspend(&DCF77Control);
      DWT->CYCCNT = 0x00000000; // nach DCF77_Suspend()-Aufruf!
      DCF77_ResumeInit(&DCF77Control);
      while(DWT->CYCCNT<SleepTime)
      {
        // die gesamte Datenverarbeitung (ADC, DMA, Regelcode, ...) ist eingefroren
      }
      DCF77_Resume(&DCF77Control);
      DWT->CTRL = 0x40000000;
    }
  }
}

//################################################### main_RunExampleByGPIOA_Pin
int main_RunExampleByGPIOA_Pin(void)
{
  void USART_PrintHelp()
  {
    // (der TDCF77...-Kram nur damit uprintf() geht)
    TDCF77State   DCF77State;
    TDCF77Control DCF77Control;
    Set_HSE(fMHz);
    DCF77_Init(&DCF77Control, &DCF77State, 0);
    while(DCF77State.SMPLoop<1) // Sleep(1s)
    {
      __WFI();
    }
    uprintf("> Zur Auswahl GPIO-Pin mit GND verbinden:\r\n\r\n");
    uprintf("  PA0 : USART_LCResonanceMeter\r\n");
    uprintf("  PA1 : LED_SignalQuality\r\n");
    uprintf("  PA2 : LED_MinuteSignal\r\n");
    uprintf("  PA3 : GPIO_SyncSignals\r\n");
    uprintf("  PA4 : USART_DCF77\r\n");
    uprintf("  PA5 : USART_DecodeBitValue\r\n");
    uprintf("  PA6 : USART_DecodeHitValue\r\n\r\n\r\n");
    DCF77_Deinit(&DCF77Control);
    USART1->CR1 = 0;
  }
  while(true)
  {
    // GPIOA0 bis GPIOA6 auf Eingang mit Pullup
    RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
    GPIOA->CRL = (GPIOA->CRL&0xf0000000)|0x08888888;
    GPIOA->BSRR = 0x7f;
    USART_PrintHelp();
    if((GPIOA->IDR&(1<<0))==0) return main0_USART_LCResonanceMeter();
    if((GPIOA->IDR&(1<<1))==0) return main1_LED_SignalQuality();
    if((GPIOA->IDR&(1<<2))==0) return main2_LED_MinuteSignal();
    if((GPIOA->IDR&(1<<3))==0) return main3_GPIO_SyncSignals();
    if((GPIOA->IDR&(1<<4))==0) return main4_USART_DCF77();
    if((GPIOA->IDR&(1<<5))==0) return main5_USART_DecodeBitValue();
    if((GPIOA->IDR&(1<<6))==0) return main6_USART_DecodeHitValue();
  }
  return 0;
}

//################################################### main
int main(void)
{
  // - USART-Einstellungen
  //   9600 Baud
  //      8 Datenbits
  //      0 Parity-Bit
  //      1 Stopp-Bit
  //   (Bluepill USART1-TXD-Ausgang ist GPIO_PA9)

  // a) einzelnes Beispiel direkt auswhlen
  // return main0_USART_LCResonanceMeter();
  // return main1_LED_SignalQuality();
  // return main2_LED_MinuteSignal();
  // return main3_GPIO_SyncSignals();
  // return main4_USART_DCF77();
  // return main5_USART_DecodeBitValue();
  // return main6_USART_DecodeHitValue();
  // return main7_SuspendResume();

  // b) Beispiele 1..7 per Kabel zwischen GPIOA und GND auswhlbar
  return main_RunExampleByGPIOA_Pin();
}
